/*
 * nono
 * Copyright (C) 2021 nono project
 * Licensed under nono-license.txt
 */

#include "optestm88k_main.h"
#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/siginfo.h>

// グローバルワーク (r2 でアセンブラに渡される)
int global_work[2];

// オーバーフロー例外なら 1、除算例外なら 2 をセットする。
// テスト前に 0 で初期化しておくこと。
#define exception_occurred (global_work[0])
// 例外を期待するなら 1 か 2、期待しないなら 0 (テスト側が書く)
#define exception_expected (global_work[1])

static void
signal_handler(int signo, siginfo_t *si, void *ctx)
{
	if (si->si_signo == SIGFPE) {
		if (si->si_code == FPE_INTOVF) {
			exception_occurred = 1;
			return;
		} else if (si->si_code == FPE_INTDIV) {
			exception_occurred = 2;
			return;
		}
		printf("%s: signo=SIGFPE si_code=%x\n", __func__, si->si_code);
	}
	if (si->si_signo == SIGILL) {
		if (si->si_code == ILL_ILLOPC) {
			exception_occurred = 0x1000;
			return;
		}
		printf("%s: signo=SIGILL si_code=%x\n", __func__, si->si_code);
	}
	// XXX
	exception_occurred = -1;
}

int
main(int ac, char *av[])
{
	char cause[64];
	struct testentry *t;
	func_t func;
	const char *arg_name = "";
	int i;
	int j;
	int num;
	int res;

	if (ac >= 2) {
		arg_name = av[1];
	}

	struct sigaction act;
	memset(&act, 0, sizeof(act));
	act.sa_sigaction = signal_handler;
	act.sa_flags = SA_SIGINFO;
	sigaction(SIGFPE, &act, NULL);
	sigaction(SIGILL, &act, NULL);

	for (i = j = 0; test_table[i].code != NULL; i++, j = 0) {
		t = &test_table[i];

		if (strncasecmp(t->testname, arg_name, strlen(arg_name)) != 0) {
			break;
		}

		if (j == 0) {
			// テスト数を数える
			for (num = 0; t->testdata[num] != NULL; num++)
				;

			printf("%s (%d) ... ", t->testname, num);
			fflush(stdout);
		}

		for (; j < num; j++) {
			exception_occurred = 0;

			res = (t->code)(global_work, t->testdata[j]);
			// 失敗なら非 0 を返す
			// bit3 は演算結果が期待と異なる
			// bit0 は例外有無が期待と異なる
			if (res != 0) {
				cause[0] = '\0';
				if ((res & 8)) {
					strlcpy(cause, "incorrect result", sizeof(cause));
				}
				if ((res & 1)) {
					if (cause[0]) {
						strlcat(cause, ", ", sizeof(cause));
					}
					if (exception_expected) {
						// 例外を期待したが、例外が起きなかった。
						strlcat(cause, "exception not occurred", sizeof(cause));
					} else {
						// 例外が起きないことを期待したのに例外が起きた。
						strlcat(cause, "exception occurred", sizeof(cause));
					}
				}
				printf("FAILED (%s) at test_%s_%d\n",
					cause, t->testname, j);
				exit(1);
			}
		}

		printf("ok\n");
	}

	return 0;
}
