//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

// AF_PACKET を用いたドライバ
// Linux 2.2 以降専用

#include "netdriver_afpacket.h"
#include "autofd.h"
#include <ifaddrs.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/if_arp.h>

// コンストラクタ
NetDriverAFPacket::NetDriverAFPacket(EthernetDevice *parent_, int loglevel_,
	const std::string& ifname_)
	: inherited("AFPacket", parent_, loglevel_)
{
	ifname = ifname_;
	if (ifname.empty()) {
		ifname = "auto";
	}
}

// デストラクタ
NetDriverAFPacket::~NetDriverAFPacket()
{
	Close();
}

// ドライバ初期化 & オープン
bool
NetDriverAFPacket::InitDriver()
{
	packet_mreq mreq {};
	int len;
	socklen_t so_len;

	putmsg(1, "trying afpacket...");
	putmsg(1, "argument: ifname=\"%s\"", ifname.c_str());

	sll.sll_family = AF_PACKET;
	sll.sll_protocol = htons(ETH_P_ALL);
	sll.sll_ifindex = SelectInterface();
	if (sll.sll_ifindex == 0) {
		goto abort;
	}

	// bind するまでの期間、このソケットにはシステムのすべての
	// インタフェースからのパケットが着信することに注意。
	fd = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (fd < 0) {
		errmsg = string_format("%s: %s", "socket(AF_PACKET)", strerror(errno));
		putmsg(1, "%s", errmsg.c_str());
		if (errno == EPERM) {
			// Linux では setcap(8) が必要。分かりづらいのでメッセージを出すが
			// その例示より先に errmsg を表示したい。
			warnx("%s", errmsg.c_str());
			errmsg.clear();

			fprintf(stderr,
				" ** Set CAP_NET_RAW capability to the nono executable file.\n"
				" **  ex) sudo setcap cap_net_raw=ep path/to/%s\n",
				getprogname());
		}
		return false;
	}

	if (bind(fd, (const sockaddr *)&sll, sizeof(sll)) != 0) {
		errmsg = string_format("bind(%s): %s", ifname.c_str(), strerror(errno));
		goto abort;
	}

	// PROMISC セット
	mreq.mr_ifindex = sll.sll_ifindex;
	mreq.mr_type = PACKET_MR_PROMISC;
	if (setsockopt(fd, SOL_PACKET, PACKET_ADD_MEMBERSHIP,
		&mreq, sizeof(mreq)) != 0) {
		errmsg = string_format("%s: %s", "set PROMISC", strerror(errno));
		goto abort;
	}

	so_len = sizeof(len);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &len, &so_len) != 0) {
		errmsg = string_format("%s: %s",
			"getsockopt(SO_RCVBUF)", strerror(errno));
		goto abort;
	}

	putmsg(1, "opened for %s", ifname.c_str());
	putmsg(1, "warning: communicating with the localhost is not possible");

	// AFPacket では ifup スクリプトは実行しない

	return true;

 abort:
	putmsg(1, "%s", errmsg.c_str());
	Close();
	return false;
}

// クローズ
// (private)
void
NetDriverAFPacket::Close()
{
	if (IsOpen()) {
		close(fd);
		fd = -1;
	}
}

// モニタ (ドライバ依存情報のみ)
void
NetDriverAFPacket::MonitorUpdateMD(TextScreen& screen)
{
	screen.Print(0, 1, "Interface: %s", ifname.c_str());
}

// パケットを送信する
bool
NetDriverAFPacket::SendPacket(const std::vector<uint8>& data)
{
	int n;

	assert(IsOpen());

	putlog(2, "SendPacket %zu bytes", data.size());
	tx_pkts++;
	tx_bytes += data.size();

	n = sendto(fd, data.data(), data.size(), 0,
		(const sockaddr *)&sll, sizeof(sll));
	if (n == -1) {
		warn("NetDriverAFPacket.SendPacket: sendto failed: fd=%d datalen=%zu",
			fd, data.size());
		return false;
	}

	if (n < (int)data.size()) {
		return false;
	}
	return true;
}

// パケットを受信する
ssize_t
NetDriverAFPacket::RecvPacket()
{
	uint8 buf[10240];
	ssize_t rv;

	rv = recv(fd, buf, sizeof(buf), 0);
	if (rv > 0) {
		RecvEnqueue(buf, rv);
	}
	return rv;
}

// インタフェース番号を返す。
// auto のときは選択したインタフェース名を ifname にセットする。
// インタフェースが決定できないときは errmsg をセットして 0 を返す。
int
NetDriverAFPacket::SelectInterface()
{
	int ifindex = 0;
	struct ifaddrs *ifa_list, *ifa;

	if (ifname == "auto") {
		// インタフェースを探す
		int level = 1;

		if (getifaddrs(&ifa_list) == -1) {
			errmsg = "getifaddrs";
			return 0;
		}

		for (ifa = ifa_list; ifa; ifa = ifa->ifa_next) {
			if ((ifa->ifa_flags & IFF_UP) == 0) {
				continue;
			}
			if (ifa->ifa_addr == NULL) {
				continue;
			}
			// AF_PACKET をサポートするインタフェースを探す
			if (ifa->ifa_addr->sa_family != AF_PACKET) {
				continue;
			}

			// ETHERNET を探す
			sockaddr_ll *sa = (sockaddr_ll *)ifa->ifa_addr;
			if (sa->sll_hatype != ARPHRD_ETHER) {
				continue;
			}

			if (ifindex == 0) {
				// 最初に見つかったものを採用
				ifname = ifa->ifa_name;
				ifindex = sa->sll_ifindex;
			} else {
				// インタフェース候補が 2個以上ある場合はログレベル 0 にしたい
				level = 0;
				break;
			}
		}
		freeifaddrs(ifa_list);

		if (ifindex == 0) {
			errmsg = "auto could not find a suitable interface";
			return ifindex;
		}

		putmsg(level, "auto-selected ifname \"%s\"", ifname.c_str());

	} else {
		ifindex = if_nametoindex(ifname.c_str());
		if (ifindex == 0) {
			errmsg = string_format("interface %s not found", ifname.c_str());
		}
	}
	return ifindex;
}
