//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// pthread の OS ごとに違う部分を吸収する

#include "mythread.h"
#include <pthread.h>
#if defined(HAVE_PTHREAD_NP_H)
#include <pthread_np.h>
#endif

// pthread_set{,_}name_np() のプロトタイプは OS によってみんなフリーダムなので
// (というか _np は Non Portable か Non Posix のことらしい?)、
// ここで吸収する。
//
// MacOSX  は void pthread_setname_np(name)
// Linux   は int  pthread_setname_np(threadID, name)
// NetBSD  は int  pthread_setname_np(threadID, name, arg)
// OpenBSD は void pthread_set_name_np(threadID, name)
// FreeBSD は void pthread_set_name_np(threadID, name)
// (2020/06 以降は pthread_setname_np(threadID, name) も追加)
void
PTHREAD_SETNAME(const char *name)
{
#if defined(__linux__)
	// Linux (Ubuntu 20.04) の top はスレッド名があればそれだけ表示するらしい。
	// スレッド名 "Main" がプロセス名なしに表示されても何のことか分からないので
	// プロセス名を入れておきたい。
	// 一方 NetBSD の top ではプロセス名とスレッド名は欄が別なので(そして狭い)
	// ここにプロセス名は入れたくない。ので、Linux の時だけに分離する。
	// 他の OS はしらん。
	char buf[32];

	snprintf(buf, sizeof(buf), "nono:%s", name);
	name = buf;
#endif

#if defined(HAVE_PTHREAD_SETNAME_NP_name)
	pthread_setname_np(name);

#elif defined(HAVE_PTHREAD_SETNAME_NP_th_name)
	pthread_setname_np(pthread_self(), name);

#elif defined(HAVE_PTHREAD_SETNAME_NP_th_name_arg)
	pthread_setname_np(pthread_self(), name, NULL);

#elif defined(HAVE_PTHREAD_SET_NAME_NP_th_name)
	pthread_set_name_np(pthread_self(), name);

#else
	// デバッガとかでスレッドに名前がついてると嬉しいという程度なので、
	// なければ気にしない。
#endif
}
