//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m88100.h"
#include "m88100excep.h"
#include "debugger.h"
#include "cvprompt.h"
#include "mpu.h"
#include "scheduler.h"

#define OP_DEF(name)	void __CONCAT(m88kcpu::op_,name)()
#define OP_FUNC(name)	__CONCAT(op_,name)()
#include "m88100ops.cpp"

// 文字列を作る前にレベルを評価するためマクロ
#define cpulog(lv, fmt...)	do {	\
	if (__predict_false(gMPU->loglevel >= (lv))) {	\
		gMPU->putlogn(fmt);	\
	}	\
} while (0)

void
m88kcpu::PowerOn()
{
	// 履歴は電源オン時だけ初期化。
	// (LUNA88K は ROM が正常パスで CPU リセットを行ったりするので)
	exhist.Clear();
	brhist.Clear();

	// この後起きるリセット例外で xip を初期化する前に参照することになるので
	// これだけここでも初期化しておく。
	xip = 0;
}

// リセット例外のようなもの?
void
m88kcpu::Reset()
{
	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	// リセット例外も発生時の XIP を記録する。(LUNA88K の ROM 1.20)
	exhist.AddEntry(xip | 1, 0, 0xfc000000 | 0);
	brhist.AddEntry(xip | 1, 0, 0xfc000000 | 0);
	// デバッガに通知 (例外ブレーク用)
	debugger_notify_exception(0);

	// 内部の割り込み状態をクリア
	intr_pending = 0;
	atomic_reqflag &= ~(CPU_REQ_RESET | CPU_REQ_INTR | CPU_REQ_STOP);

	// Table 6-8 (p6-47)
	psr = 0x800003ff;
	// XXX ただし SFU1 は未実装なので下げておく
	psr &= ~PSR_SFD1;

#if defined(TEST_INTERRUPT)
	psr &= ~PSR_IND;
	psr &= ~PSR_SFRZ;
#endif

	SetPSR();

	fip = 0;

	// レジスタのうち明記されてるものをクリア。
	// 不定と明記されてるものはリセットでは触らないでおく。
	// scoreboard: cleared
	nip = 0;
	xip = 0;
	vbr = 0;
	fpecr = 0;
	fpcr = 0;
	fpsr = 0;
	// DMTx は bit0(Valid) をクリア。他は不定
	dmt0 &= ~DM_VALID;
	dmt1 &= ~DM_VALID;
	dmt2 &= ~DM_VALID;

	cmmu[0].Reset();
	cmmu[1].Reset();

	fetch();
}

// 仮想時間 delta [nsec] だけ CPU を実行する。実際には大抵行き過ぎる。
// 戻り値は CPU 状態 (Scheduler::SCHED_CPU_*)。
uint32
m88kcpu::Run(uint32 delta)
{
	bool is_release = false;
	bool is_intr = false;

	assert(used_cycle == 0);

	goal_cycle = Vtime2Cycle(delta);
	if (__predict_false(goal_cycle == 0)) {
		goal_cycle = 1;
	}

	if ((atomic_reqflag & CPU_REQ_STOP)) {
		if ((atomic_reqflag & (CPU_REQ_INTR | CPU_REQ_PROMPT | CPU_REQ_RESET))) {
			// 割り込み、デバッガ、リセットが来たら STOP 解除。
			atomic_reqflag &= ~CPU_REQ_STOP;
		} else {
			// STOP 継続として戻る。
			used_cycle = goal_cycle;
			goto exit;
		}
	}

	while (!is_release) {
		// shift pipeline
		opX = opF;
		xip = nip;

		// prefetch next inst
		fetch();

		// この時点で実行しようとする命令が opX に入っている

		if (atomic_reqflag != 0) {
			// 何かが起きた

			if ((atomic_reqflag & CPU_REQ_RESET)) {
				// リセット例外。
				Reset();
				// XIP がリセットベクタを指すためにはもう一回フェッチが必要
				// なので、ここは continue が必要。
				continue;
			}

			// 割り込みの認識
			// 割り込みの実行は命令実行完了後、次のプリフェッチの前
			if ((atomic_reqflag & CPU_REQ_INTR)) {
				atomic_reqflag &= ~CPU_REQ_INTR;
				if (intr_pending && IsIntrEnable()) {
					// この命令のあとに割り込み処理
					is_intr = true;
				}
#if defined(TEST_INTERRUPT)
				atomic_reqflag |= CPU_REQ_PROMPT;
#endif
			}

			if ((atomic_reqflag & CPU_REQ_TRACE)) {
				// デバッガプロンプトを出すかどうかを判断する
				//inst_count++;
				if (debugger_check()) {
					atomic_reqflag |= CPU_REQ_PROMPT;
				}
			}

			if ((atomic_reqflag & CPU_REQ_PROMPT)) {
				// デバッガプロンプトのための実行停止

				// プロンプト獲得を通知
				gCVPrompt->NotifyAcquire();

				// プロンプト解放を待機
				gCVPrompt->WaitRelease();

				// プロンプトを抜けたのでここでフラグを落とす
				atomic_reqflag &= ~CPU_REQ_PROMPT;

				// プロンプトから戻ってきたら一旦スケジューラまで戻る(?)。
				// reset コマンドのようにデバッガからスケジューラに出した
				// リクエストを(なるはやで?)回収するため。
				atomic_reqflag |= CPU_REQ_RELEASE;
			}

			if ((atomic_reqflag & CPU_REQ_RELEASE)) {
				// 中断リクエスト受信
				// この命令を実行し終わったら抜ける
				atomic_reqflag &= ~CPU_REQ_RELEASE;
				is_release = true;
			}
		}

		if (OpIsBusErr(opX)) {
			// プリフェッチしていた命令がバスエラー
			Exception(M88K_EXCEP_INST);
			continue;
		}

		uint32 op12 = op32_to_12(opX);

		switch (op12) {
#include "m88100switch.inc"
		 default:
			OP_FUNC(illegal);
			break;
		}

		if (is_intr && IsIntrEnable()) {
			// 割り込み処理要求が来ていて、直前の命令が割り込みを禁止していなければ割り込み
			// XXX 実際は違う気がする
			is_intr = false;
			cpulog(4, "INTR take");
			ExceptionCore(M88K_EXCEP_INTERRUPT, ExceptionKind::INTR);
			continue;
		}

		used_cycle++;

		if ((atomic_reqflag & CPU_REQ_STOP)) {
			// STOP 状態は、残りサイクルを消費したことにして戻る
			cpulog(2, "STOP 状態検出");
			if (used_cycle < goal_cycle) {
				used_cycle = goal_cycle;
			}
		}

		// XXX 命令の実行でリリースするのに、いったんループを回って
		// からではタイミングが遅すぎる
		if ((atomic_reqflag & CPU_REQ_RELEASE)) {
			// 中断リクエスト受信
			// この命令を実行し終わったら抜ける
			atomic_reqflag &= ~CPU_REQ_RELEASE;
			is_release = true;
		}

		if (used_cycle >= goal_cycle)
			break;
	}

 exit:
	total_vtime += Cycle2Vtime(used_cycle);
	used_cycle = 0;

	// HALT 状態がないので NORMAL か STOP かだけでいい
	return (atomic_reqflag & CPU_REQ_STOP)
		? Scheduler::SCHED_CPU_STOP
		: Scheduler::SCHED_CPU_NORMAL;
}

// この命令後に実行を中断して VM に戻る
void
m88kcpu::Release()
{
	atomic_reqflag |= CPU_REQ_RELEASE;
}

// リセット例外を発行する
void
m88kcpu::RequestReset()
{
	atomic_reqflag |= CPU_REQ_RESET;
}

// 通常の例外。
// DataAccessException, エラー、割り込み、TRAP からは使わない。
// vec はベクタ番号。
void
m88kcpu::Exception(int vec)
{
	ExceptionCore(vec, ExceptionKind::NORMAL);
}

// 例外処理コア部分
// vec はベクタ番号。
void
m88kcpu::ExceptionCore(int vec, ExceptionKind cause)
{
	uint syscall = 0;

	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	// OpenBSD システムコールなら番号もついでに記録。
	// システムコール番号は tb0 発行時点で r13 にセットされている。
	// ただし 0 番は syscall(2) (間接システムコール) で、実際の番号は r2。
	// 198番 __syscall(2) だと r2:r3 (なので実質 r3)。
	if (vec == 450) {
		syscall = r[13];
		if (syscall == 0) {
			syscall = r[2];
		} else if (syscall == 198) {
			syscall = r[3];
		}
		syscall &= 0xfff;
	}
	uint32 from = xip | (IsSuper() ? 1 : 0);
	uint32 inst = 0xfc000000 | (syscall << 12) | vec;
	exhist.AddEntry(from, 0, inst);
	brhist.AddEntry(from, 0, inst);
	// デバッガに通知 (例外ブレーク用)
	debugger_notify_exception(vec);

	// 命令中で起きる例外はとりあえず全部2クロック追加? (Table.7-5)
	if (cause != ExceptionKind::INTR) {
		AddCycle(2);
	}

	if (psr & PSR_SFRZ) {
		if (cause != ExceptionKind::TRAP) {
			cpulog(2, "SFRZ Error Exception");
			vec = M88K_EXCEP_ERROR;
			cause = ExceptionKind::ERROR;
		}
	} else {
		cpulog(2, "Exception %x", vec);
		if (vec != 1) {
			cpulog(1, "Exception %x x=%x n=%x f=%x", vec, xip, nip, fip);
		}
		sxip = xip;
		snip = nip;
		sfip = fip;

		sxip |= SIP_V;
		snip |= SIP_V;
		sfip |= SIP_V;

		if (OpIsBusErr(opX)) {
			sxip |= SIP_E;	// E bit
		}
		if (OpIsBusErr(opF)) {
			snip |= SIP_E;	// E bit
		}

		if (vec == M88K_EXCEP_UNIMPL_OP) {
			sxip &= ~SIP_V;
		}
	}

	if (cause != ExceptionKind::DATA) {
		// データアクセス例外以外は、
		// データパイプラインはフラッシュ状態とする。
		// ほんとうは違う。かも。
		dmt0 = 0;
	} else {
		// xip: ld/st/xmem
		// nip: next
		// fip: fetch

		// rom1.2 は DAE が起きると、

		// ROM挙動               CPU の RTE 挙動
		// sfip = sxip
		// sxip -> ~V (INVALID)  CPU はそもそも sxip を評価しない
		// snip = 0 (INVALID)    CPU はこれを nop 相当に処理する
		// rte

		// してくる。そのまま読むと命令を再実行しようとしているように
		// 読めるが、再実行すると無限ループになってしまう。
		// DAE の時点で xip が進んでいることを期待して利用している?

		// openbsd は
		// xmem の write バスエラーを検知すると命令を再実行するために
		// fip = nip, nip = xip して rte する。ということは xip = xmem に
		// なっていることを期待している。

		// というわけで
		// ld で DataAccessException のときは、実CPUはすでに次の命令を
		// 実行している?
		// もう少し考えると、スコアボードでレジスタの干渉がなければ、
		// 実行している?

		// 仕方がないのでハックでごまかす。
		// rte のほうも参照のこと。

		if (IsSuper() && xip >= 0x41000000 && xip < 0x42000000) {
			// 遅延例外機構がないので、ROM 合わせのハック。
			sxip = snip;
		}
	}

	epsr = psr;
	SetPSR(psr | PSR_SUPER | PSR_SFD1 | PSR_IND | PSR_SFRZ);

	fip = vbr + (vec << 3);
	fetch();
	// ここでベクタに飛ぶのだが、ベクタ内の2命令でおそらく必ずもう一度
	// 分岐してそっちでも履歴が残るので、ここでブランチ履歴を残すのは
	// ちょっと冗長という気もする。

	if (OpIsBusErr(opF)) {
		if (cause == ExceptionKind::ERROR) {
			// double bus fault
			// 実機は無限ループ状態
			PANIC("Infinite ERROR (double bus fault)");
		} else {
			ExceptionCore(M88K_EXCEP_ERROR, ExceptionKind::ERROR);
		}
	}
}

inline
uint32
m88kcpu::Calcdmt(uint32 flag)
{
	return
		   ((psr & PSR_BO_LE) ? DM_BO : 0)
	     | ((IsSuper() && !(flag & DM_USR)) ? DM_DAS : 0)
	     | DM_VALID;
}

// データアクセス例外(Read)
// ld.b, ld.h, ld.hu, ld 用
void
m88kcpu::ReadDataException32(uint32 addr, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr & ~0x3;
	dmt0 = Calcdmt(flag)
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	;
	dmt1 &= ~DM_VALID;
	dmt2 &= ~DM_VALID;
	cpulog(1, "ReadDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(Read)
// ld.d 用
void
m88kcpu::ReadDataException64(uint32 addr, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr;
	dmt0 = Calcdmt(flag)
	     | DM_DOUB1			// ダブルワードの第1アクセスのときセット
	     | (FLD_D << 7)
	     | DM_EN_MASK		// 32ビット有効
	;

	dma1 = dma0 + 4;
	dmt1 = Calcdmt(flag)
	     | (FLD_D2 << 7)
	     | DM_EN_MASK			// 32ビット有効
	;

	dmt2 &= ~DM_VALID;

	if ((flag & DM_DOUB1) == 0) {
		// .d の2ワード目がバスエラー。第1ワードのアクセスは成功した。
		dmt0 &= ~DM_VALID;
	}

	cpulog(1, "ReadDAE64 xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(Write)
// st.b, st.h, st.hu, st 用
void
m88kcpu::WriteDataException32(uint32 addr, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr & ~0x3;
	dmt0 = Calcdmt(flag)
	     | (FLD_D << 7)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | DM_WRITE
	;
	dmd0 = rD;

	dmt1 &= ~DM_VALID;
	dmt2 &= ~DM_VALID;
	cpulog(1, "WriteDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(Write)
// st.d 用
void
m88kcpu::WriteDataException64(uint32 addr, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr & ~0x3;
	dmt0 = Calcdmt(flag)
	     | DM_DOUB1		// ダブルワードの第1アクセスのときセット
	     | (FLD_D << 7)
	     | DM_EN_MASK
	     | DM_WRITE
	;
	dmd0 = rD;

	dma1 = dma0 + 4;
	dmt1 = Calcdmt(flag)
	     | (FLD_D2 << 7)
	     | DM_EN_MASK
	     | DM_WRITE
	;
	dmd1 = rD2;

	dmt2 &= ~DM_VALID;

	if ((flag & DM_DOUB1) == 0) {
		// .d の2ワード目がバスエラー。第1ワードのアクセスは成功した。
		dmt0 &= ~DM_VALID;
	}

	cpulog(1, "WriteDAE64 xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(xmem)
// xmem には .d は無い
void
m88kcpu::XmemDataException(uint32 addr, uint32 flag)
{
	dma0 = addr & ~0x3;
	dmt0 = Calcdmt(flag)
	     | DM_LOCK		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	;

	dma1 = dma0;
	dmt1 = Calcdmt(flag)
	     | DM_LOCK		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | DM_WRITE
	;
	dmd1 = rD;

	dmt2 &= ~DM_VALID;

	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	if (cmmu[1].fault_code != m88200::FAULT_CODE_BUSERR
	 || cmmu[1].acc_read) {
		// アドレス変換フォルトまたは read 時点
		// nop
	} else {
		// xmem の write 時点
		dmt0 &= ~DM_VALID;
	}

	cpulog(1, "xmemDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// FP 例外
// fpex: FP例外コード M88K_FPEXCEP
void
m88kcpu::FPException(int fpex)
{
	// FDVZ 以上の例外が PRECISE。
	// 未満の FUNF, FOVF, FINX は IMPRECISE。
	int vec = (fpex >= M88K_FPEXCEP_FDVZ) ? M88K_EXCEP_SFU1_PRECISE : M88K_EXCEP_SFU1_IMPRECISE;
	fpecr = fpex;

	Exception(vec);
}

void
m88kcpu::Interrupt(int level)
{
	// level は 0 か 1。
	intr_pending = level;
	atomic_reqflag |= CPU_REQ_INTR;
}

void
m88kcpu::fpu_unimpl()
{
	printf("%s\n", __func__);
	FPException(M88K_FPEXCEP_FUNIMP);
}

OP_DEF(illegal)
{
	// 本来は不当命令例外?
	cpulog(0, "Illegal instruction %08x", (uint32)opX);
}
