//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "object.h"
#include <vector>

#if BYTE_ORDER == LITTLE_ENDIAN
#define HB(addr)	((addr) ^ 1)
#define HLB(addr)	((addr) ^ 3)
#define HLW(addr)	((addr) ^ 2)
#else
#define HB(addr)	(addr)
#define HLB(addr)	(addr)
#define HLW(addr)	(addr)
#endif

class Device : public Object
{
	using inherited = Object;
 protected:
	Device(const std::string& objname_);
 public:
	virtual ~Device() override;

 public:
	// 初期化は以下の手順で行う。
	// 1. コンストラクタ
	//	メンバ変数の初期値代入などを行う。失敗するケースは考慮していない。
	//	gConfig の内部変数の追加はここ(か2.の動的コンストラクション)で
	//	行うこと。この時点ですでに gConfig の設定内容の参照は可能だが、
	//	エラーが出るものはここではなく Init() で処理すること。
	//	putlog() は使える。厳密には MPU コンストラクタ以降で使えるが
	//	MPU より前は BusErrDevice しかおらずこいつは putlog() しないので。
	//
	// 2. Create() ... 動的コンストラクション
	//	必要なら設定ファイルによって可変なオブジェクトのコンストラクトを行う。
	//	例えば SCSI ドライブなどが該当する。
	//	ここでも gConfig の参照および内部変数の登録は可能。
	//
	// 3. parse_logopt() ... -L オプションの処理
	//	この時点で Object 由来のオブジェクトはすべて出揃っている必要がある。
	//	またここで --show-config も行うため、この時点で gConfig の内容は確定し
	//	以降 gConfig の変更は出来ない。
	//
	// 3.1. InitDriver() ... NetDriver を選択
	//	NetDriver の選択はログが準備出来た後で行いたいが、スレッドの動く
	//	Init() よりは前で行いたいのでここで特別扱いしている。
	//
	// 3.2. ParseMonitors() ... -M オプションの処理
	//	(GUI 版のみ) この時点で Monitor がすべて登録されている必要がある。
	//
	// 4. Init() ... 初期化
	//	ここでは ROM などのファイルの読み込み、またスレッドの起動などを行う。
	//	Object 由来のオブジェクトはすべて揃っているため、相互にアクセス可能。
	//	ただしその際の依存関係については、vm.cpp (や xiospace.cpp) 内の
	//	コンストラクタの順序で調整すること (循環参照不可)。
	//
	// 5. Apply() ... 動的設定の適用
	//	実行中に変更可能なパラメータもここで一度適用する。
	//
	// 6. PowerOn(), ResetHard() ... VM 起動

	// 動的コンストラクション。
	// この時点で設定ファイルは読める。putlog() は使えない。
	// 成功すれば true を返す、
	// 失敗なら warn()/warnx() でエラーメッセージを表示して false を返すこと。
	virtual bool Create();

	// 初期化。
	// 起動時設定の適用、ファイルの読み込み、スレッド作成を行う。
	// Object 由来のオブジェクトはすべて揃っており、設定ファイルは読める。
	// putlog() も使える。
	// 成功すれば true を返すこと。
	// 失敗なら warn()/warnx()、あるいは ConfigItem::Err() 系 (実態は warnx)
	// でエラーメッセージを表示して false を返すこと。
	virtual bool Init();

	// 現在の設定を VM に反映させる。
	// ここで適用するのは実行中に変更可能なパラメータのみ。
	// 適用内容を表示したい時は putlog(1) が妥当か。
	// 成功すれば true を返すこと。
	// 失敗なら warn()/warnx()、あるいは gConfig->Err() 系 (実態は warnx) で
	// エラーメッセージを表示して false を返すこと。
	// といっても false が返されたらどうするんだろうか。
	virtual bool Apply();

	virtual bool PowerOn();
	virtual bool PowerOff();

	// ResetHard() はこのデバイスのリセット。
	// このデバイスがどのタイミングでリセットされるかは VM による。
	// X68030 なら vm_x68k.cpp 内のコメント参照。
	// LUNA は詳細不明。
	virtual void ResetHard();
};

class IODevice : public Device
{
	using inherited = Device;
 protected:
	IODevice(const std::string& objname_);
 public:
	virtual ~IODevice() override;

 public:
	// アクセス関数。
	// Read*() は読み込んだ値を (uint32) で返す。
	// Write*() は書き込めたら 0 を返す。
	// Peek8() はデバッガアクセスで読み込んだ値を (uint32) で返す。
	// いずれもバスエラーなら (uint64)-1 を返す。
	// (signed) int64 を使ってないのはノーマルパスで signed/unsigned の
	// キャスト事故を防ぐため。
	virtual uint64 Read8(uint32 addr);
	virtual uint64 Read16(uint32 addr);
	virtual uint64 Read32(uint32 addr);
	virtual uint64 Write8(uint32 addr, uint32 data);
	virtual uint64 Write16(uint32 addr, uint32 data);
	virtual uint64 Write32(uint32 addr, uint32 data);
	virtual uint64 Peek8(uint32 addr);

	uint32 devaddr {};			// このデバイスの先頭アドレス

	// (必要なら)このデバイスの占めるバイト数。
	// 今のところ XIOSpace、MuxDevice では必要。それ以外では不要。
	uint32 devlen {};
};

class BusErrDevice : public IODevice
{
	using inherited = IODevice;
 public:
	BusErrDevice();
	virtual ~BusErrDevice() override;

 public:
	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;
};

// 何もしないデバイス。
// 読み込むと全域で 0xff が読めて、書き込んでも何も起きない。
class NopIODevice : public IODevice
{
	using inherited = IODevice;
 public:
	NopIODevice();
	virtual ~NopIODevice() override;

 public:
	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;
};

// 前方参照
class DMACDevice;
class Human68k;
class InterruptDevice;
class LunafbDevice;
class MK48T02Device;
class MPUDevice;
class SCSIDevice;
class SCSIDisk;
class SCSITarget;

extern std::vector<Device *> gDevices;
extern std::unique_ptr<BusErrDevice> gBusErr;
extern std::unique_ptr<NopIODevice> gNopIO;
