//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "interrupt.h"
#include "dmac.h"
#include "lance.h"
#include "pedec.h"
#include "m68030.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "scc.h"
#include "sio.h"
#include "spc.h"
#include "sysclk.h"

//                           IODevice
//                              |
//                              v
//                       InterruptDevice
//                              |
//     +------------------------+-------------------------+
//     |                        |                         |
//     |                        v                         |
//     |                  M680x0Interrupt                 |
//     |                        |                         |
//     |               +--------+--------+                |
//     v               v                 v                v
// PEDECDevice    X68030Interrupt    LunaInterrupt   SysCtlrDevice
// (X68030のLv1   (X68030)           (LUNA-I)        (LUNA88K)
//  カスケード)

std::unique_ptr<InterruptDevice> gInterrupt;

//
// 仮想割り込みコントローラ (共通部)
//

// コンストラクタ(オブジェクト名指定あり)
InterruptDevice::InterruptDevice(const std::string& objname_)
	: inherited(objname_)
{
}

// コンストラクタ(オブジェクト名指定なし)
InterruptDevice::InterruptDevice()
	: InterruptDevice("Interrupt")
{
}

// デストラクタ
InterruptDevice::~InterruptDevice()
{
}


//
// 仮想割り込みコントローラ (m680x0 共通部)
//

// CPU コアからの割り込みアクノリッジを受け付けるグローバル関数
int
m68030_interrupt_acknowledge(int lv)
{
	return gInterrupt->InterruptAcknowledge(lv);
}

// コンストラクタ
M680x0Interrupt::M680x0Interrupt()
{
}

// デストラクタ
M680x0Interrupt::~M680x0Interrupt()
{
}

// リセット
void
M680x0Interrupt::ResetHard()
{
	ipl = 0;
	intmap = 0x00000001;
	memset(&counter, 0, sizeof(counter));
}

// 子デバイスが割り込み信号線をアサートした (m680x0 共通)
void
M680x0Interrupt::AssertINT(Device *source)
{
	uint32 oldmap;
	uint32 srcmap;
	int newipl;

	oldmap = intmap;
	srcmap = GetIntmap(source);
	intmap |= srcmap;

	if ((oldmap ^ intmap) != 0) {
		// 立ち上がりエッジでカウント
		counter[__builtin_clz(srcmap)]++;
	}

	newipl = (uint)(31 - __builtin_clz(intmap)) / 4;

	// アサートによって割り込みレベルが上がったら、新しいレベルを通知
	// (現在通知中より低い割り込みは受け付けただけでここで保留)
	if (newipl > ipl) {
		ipl = newipl;
		gMPU->Interrupt(ipl);
	}
}

// 子デバイスが割り込み信号線をネゲートした (m680x0 共通)
void
M680x0Interrupt::NegateINT(Device *source)
{
	int newipl;

	intmap &= ~GetIntmap(source);
	// intmap の最下位ビットは常に立ててあり intmap は 0 にならない
	newipl = (uint)(31 - __builtin_clz(intmap)) / 4;

	// ネゲートによって割り込みレベルが下がったら、新しいレベルを通知
	// (以前に保留されていた低いレベルがあればここで見えるようになる)
	if (newipl < ipl) {
		ipl = newipl;
		gMPU->Interrupt(ipl);
	}
}


//
// X68030 仮想割り込みコントローラ
//

// コンストラクタ
X68030Interrupt::X68030Interrupt()
{
	monitor.func = (MonitorCallback_t)&X68030Interrupt::MonitorUpdate;
	monitor.SetSize(24, 11);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
X68030Interrupt::~X68030Interrupt()
{
}

// デバイスから Intmap 値を引く
uint32
X68030Interrupt::GetIntmap(Device *source) const
{
	if (__predict_true(source == gMFP.get())) {
		return IntmapMFP;
	}
	if (__predict_true(source == gSCC.get())) {
		return IntmapSCC;
	}
	if (__predict_true(source == gDMAC.get())) {
		return IntmapDMAC;
	}
	if (__predict_true(source == gPEDEC.get())) {
		return IntmapPEDEC;
	}
	if (source == NULL) {
		// XXX デバイスまだいないのでとりあえず
		return IntmapNMI;
	}
	PANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジに応答する
int
X68030Interrupt::InterruptAcknowledge(int lv)
{
	switch (lv) {
	 case 6:
		return gMFP->InterruptAcknowledge();

	 case 5:
		return gSCC->InterruptAcknowledge();

	 case 3:
		return gDMAC->InterruptAcknowledge();

	 case 1:
		return gPEDEC->InterruptAcknowledge(lv);

	 default:
		break;
	}
	PANIC("Unknown interrupt acknowledge lv=%d", lv);
}

void
X68030Interrupt::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "NMI",	IntmapNMI },
		{ "MFP",	IntmapMFP },
		{ "SCC",	IntmapSCC },
		{ "DMAC",	IntmapDMAC },
		{ "PEDEC",	IntmapPEDEC },
	}, tablePEDEC[] = {
		{ "SPC",	PEDECDevice::IntmapSPC },
		{ "FDC",	PEDECDevice::IntmapFDC },
		{ "FDD",	PEDECDevice::IntmapFDD },
		{ "HDD",	PEDECDevice::IntmapHDD },
		{ "PRT",	PEDECDevice::IntmapPRT },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	uint intr_level = gMPU680x0->GetCPU()->reg.intr_level;
	// PEDEC 分も一緒に表示したい
	uint32 pintr = gPEDEC->GetInt();

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "Lv Device    Count  IM=%d", intr_level);
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%10u", counter[clz]);
		if (lv <= intr_level) {
			screen.Puts(20, y, "Mask");
		}
		y++;
	}

	// pintr は下位ワードが割り込み状態、上位ワードが許可状態。
	uint32 pmask = pintr >> 16;
	for (int i = 0; i < countof(tablePEDEC); i++) {
		uint32 map = tablePEDEC[i].map;
		screen.Puts(3, y, TA::OnOff(pintr & map), tablePEDEC[i].name);
		// XXX カウンタはまだない
		if (intr_level >= 1 || (pmask & map) == 0) {
			screen.Puts(20, y, "Mask");
		}
		y++;
	}
}

//
// LUNA-I 仮想割り込みコントローラ
//

// コンストラクタ
LunaInterrupt::LunaInterrupt()
{
	monitor.func = (MonitorCallback_t)&LunaInterrupt::MonitorUpdate;
	monitor.SetSize(24, 5);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
LunaInterrupt::~LunaInterrupt()
{
}


// デバイスから Intmap 値を引く
uint32
LunaInterrupt::GetIntmap(Device *source) const
{
	if (__predict_true(source == gSIO.get())) {
		return IntmapSIO;
	}
	if (__predict_true(source == gSysClk.get())) {
		return IntmapSysClk;
	}
	if (__predict_true(source == gEthernet.get())) {
		return IntmapLance;
	}
	if (__predict_true(source == gSPC.get())) {
		return IntmapSPC;
	}
	PANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジに応答する
int
LunaInterrupt::InterruptAcknowledge(int lv)
{
	// LUNA は全てオートベクタ
	return AutoVector;
}

void
LunaInterrupt::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "SIO",	IntmapSIO },
		{ "Clock",	IntmapSysClk },
		{ "Lance",	IntmapLance },
		{ "SPC",	IntmapSPC },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	uint intr_level = gMPU680x0->GetCPU()->reg.intr_level;

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "Lv Device    Count  IM=%d", intr_level);
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%10u", counter[clz]);
		if (lv <= intr_level) {
			screen.Puts(20, y, "Mask");
		}
		y++;
	}
}
