//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "mk48t02.h"
#include "bitops.h"
#include "mainapp.h"
#include "mpu.h"

//
// MK48T02
//
// SRAM 部分はビッグエンディアンのままメモリに置かれるので、
// 必要ならアクセス時に都度変換すること。
//
// 仕様は 0x45000000 から 2KB だが、
// 0x44000000 から 0x48000000 の手前まで 2KB 単位でミラーが見える。

// 繰り上がりについてのメモ。
//
// PROM で以下のようにすれば確認できる。
//  >450007f8;l,80595923
//  >450007fc;l,01DDMMYY
//  >450007f8;,0
//  >450007f8
//
// 2月の日数:
//  2/30 -> 03/01
//  2/31 -> 03/01
//
// 月ごとの日数:
//  01/00/30 -> 01/01/01 (00月は小)
//  01/0a/30 -> 01/0a/31 (0a月は大)
//  01/0b/30 -> 01/14/01 (0b月は小)
//  01/0c/30 -> 01/0c/31 (0c月は大)
//  01/0d/30 -> 01/14/01 (0d月は小)
//  01/0e/30 -> 01/0e/31 (0e月は大)
//  01/0f/30 -> 01/18/01 (0f月は小)
//  00/13/30 -> 01/00/01 (13月は小)
//  01/14/30 -> 01/14/31 (14月は大)
//  01/15/30 -> 01/16/01 (15月は小)
//  01/16/30 -> 01/16/31 (16月は大)
//  01/17/30 -> 02/08/01 (17月は小)
//  01/18/30 -> 01/18/31 (18月は大)
//  01/19/30 -> 01/10/01 (19月は小)
//  01/1a/30 -> 01/1a/31 (1a月は大)
//  01/1b/30 -> 01/10/01 (1b月は小)
//  01/1c/30 -> 01/1c/31 (1c月は大)
//  01/1d/30 -> 01/14/01 (1d月は小)
//  01/1e/30 -> 01/1e/31 (1e月は大)
//  01/1f/30 -> 01/18/01 (1f月は小)
//
// 月の繰り上がり後
//  01/0a/31 -> 01/0b/01
//  01/0b/31 -> 01/14/01
//  01/0c/31 -> 01/0d/01
//  01/0d/31 -> 01/14/01
//  01/0e/31 -> 01/0f/01
//  01/0f/31 -> 01/18/01
//  00/13/30 -> 01/00/01	+Y
//  01/14/31 -> 01/15/01
//  01/15/31 -> 01/16/01
//  01/16/31 -> 02/01/01	+Y
//  01/17/31 -> 02/08/01	+Y
//  01/18/31 -> 01/19/01
//  01/19/31 -> 01/10/01
//  01/1a/31 -> 02/09/01	+Y
//  01/1b/31 -> 01/10/01
//  01/1c/31 -> 01/1d/01
//  01/1d/31 -> 01/14/01
//  01/1e/31 -> 02/09/01	+Y
//  01/1f/31 -> 01/18/01
//
// 日の繰り上がり(一部)
//  01/01/0f -> 01/01/18
//  01/01/3e -> 01/02/1f
//  01/01/3f -> 01/01/2f
//
//  01/0a/28 -> 01/0b/01
//  00/0a/28 -> 00/0a/29 (閏)
//  00/0a/29 -> 00/0b/01
//  00/0a/30 -> 00/0a/31
//  00/0a/31 -> 00/0b/01
//  00/0a/32 -> 00/0a/33
//  00/0a/33 -> 00/0b/03
//  00/0a/34 -> 00/0a/35
//  00/0a/35 -> 00/0b/05
//  00/0a/36 -> 00/0a/37
//  00/0a/37 -> 00/0b/0f
//  00/0a/38 -> 00/0b/19
//  00/0a/39 -> 00/0a/21
//  00/0b/34 -> 00/14/05
//  00/0c/28 -> 00/0c/29
//  00/0c/29 -> 00/0c/30
//  00/0c/30 -> 00/0c/31
//  00/0c/31 -> 00/0d/01
//  00/0c/37 -> 00/0d/0f
//  01/1a/28 -> 01/1a/29
//
// 0a 月は閏月として判定されるようだ。bit 0?010 でマッチかな

MK48T02Device::MK48T02Device()
	: inherited("MK48T02")
{
	AddAlias("NVRAM");
	AddAlias("RTC");

	devaddr = 0x45000000;

	// 2021/07 時点の NetBSD/luna68k, OpenBSD/luna88k 合わせ。
	// MK48T02 本来定義は 0 というか 1900 というか。
	// see DecodeYear()
	year_epoch = 1970;
	use_localtime = false;

	monitor.func = (MonitorCallback_t)&MK48T02Device::MonitorUpdate;
	monitor.SetSize(40, 13);
	monitor.Regist(ID_MONITOR_MK48T02);
}

MK48T02Device::~MK48T02Device()
{
}

bool
MK48T02Device::Init()
{
	if (inherited::Init() == false) {
		return false;
	}

	reg.ex = reg.in;

	// MK48T02 のデータシートには年には西暦年の下位2桁を設定するように
	// 指示されている。
	// しかし CMU Mach luna88k, NetBSD/luna68k, OpenBSD/luna88k はいずれも
	// 西暦年から 1970 引いた値を設定するようになっている (2021/07 現在)。
	// このため、うるう年が2年ずれた状態になる。
	//
	// NetBSD で MK48T02 を採用している luna68k 以外の port は
	// ソースコード上の変数 year0 もしくは YEAR0 に
	//  newsmips, news68k: 1900 (データシート規定通り)
	//  sun3, sparc, mvme68k: 1968 (1968 はうるう年なので問題ない)
	// を採用しており、バグっていない。
	// この year0 は UNIX EPOCH とは直接関係がない RTC ハードウェアに
	// 依存する値なのだが、LUNA では混同したのだろうか?
	//
	// たぶんこのバグはもう修正されることはないだろうから、こちらで補正する
	// スイッチを用意した。

	int adjust_misused_epoch;
	adjust_misused_epoch = gConfig->Find("luna-adjust-misused-epoch").AsInt();
	if (adjust_misused_epoch) {
		// エポック年のうるう年カウンタ相当を補正項として持つ
		adjust_leap = year_epoch % 4;
	} else {
		adjust_leap = 0;
	}

	// ファイルのロード
	filename = gMainApp.GetVMDir() + "NVRAM.DAT";
	file.SetFilename(filename);
	file.SetDispname(string_format("NVRAM \"%s\"", filename.c_str()));
	mem = file.OpenCreate(2040);
	if (mem == NULL) {
		return false;
	}

	return true;
}

// バイナリ表現を BCD 表現に変換する
/*static*/ inline uint8
MK48T02Device::num2BCD(uint8 num)
{
	return ((num / 10) << 4) + (num % 10);
}

// BCD 表現をバイナリ表現に変換する
/*static*/ inline uint8
MK48T02Device::BCD2num(uint8 bcd)
{
	return (bcd >> 4) * 10 + (bcd & 0x0f);
}

uint64
MK48T02Device::Read(uint32 offset)
{
	uint32 data;

	if (offset < 2040) {
		data = mem[offset];
		putlog(3, "$%08x -> $%02x", gMPU->GetPaddr(), data);
	} else {
		switch (offset - 2040) {
		 case 0:	// コントロールレジスタ
			data = reg.ctrl;
			break;
		 case 1:	// 秒
			data = reg.ex.sec | reg.stop;
			break;
		 case 2:	// 分
			data = reg.ex.min;
			break;
		 case 3:	// 時
			data = reg.ex.hour | reg.kickstart;
			break;
		 case 4:	// 曜日
			data = reg.ex.wday | reg.freqtest;
			break;
		 case 5:	// 日
			data = reg.ex.mday;
			break;
		 case 6:	// 月
			data = reg.ex.mon;
			break;
		 case 7:	// 年
			data = reg.ex.year;
			break;
		 default:
			__unreachable();
		}

		putlog(3, "$%08x -> $%02x", gMPU->GetPaddr(), data);
	}
	return data;
}

uint64
MK48T02Device::Write(uint32 offset, uint32 data)
{
	if (offset < 2040) {
		putlog(3, "$%08x <- $%02x", gMPU->GetPaddr(), data);
		mem[offset] = data;
	} else {
		putlog(3, "$%08x <- $%02x", gMPU->GetPaddr(), data);

		switch (offset - 2040) {
		 case 0:	// コントロール
			WriteCtrl(data);
			break;
		 case 1:	// 秒 (b7 は STOP BIT)
			reg.stop = data & MK48T02Clock::STOP;
			reg.ex.sec = data & 0x7f;
			break;
		 case 2:	// 分
			reg.ex.min = data & 0x7f;
			break;
		 case 3:	// 時 (b7 は KICK START)
			reg.kickstart = data & MK48T02Clock::KICKSTART;
			reg.ex.hour = data & 0x3f;
			break;
		 case 4:	// 曜日 (b6 は Frequency Test (未実装))
			// データシートでは b7, b5, b4 は '0' とあるが実際には '1' が
			// 書き込める。書き込んだ '1' が内部クロックの更新でどうなるか
			// 未確認のため、とりあえず放置。
			reg.freqtest = data & MK48T02Clock::FREQTEST;
			reg.ex.wday = data & 0x07;
			break;
		 case 5:	// 日
			// データシートでは b7, b6 は '0' とあるが実際には b7 には '1' が
			// 書き込める (b6 には書き込めない)。b7 に書き込んだ '1' が内部
			// クロックの更新でどうなるか未確認のため、とりあえずは放置。
			reg.ex.mday = data & 0x3f;
			break;
		 case 6:	// 月
			reg.ex.mon = data & 0x1f;
			break;
		 case 7:	// 年
			reg.ex.year = data;
			break;
		 default:
			__unreachable();
		}
	}
	return 0;
}

uint64
MK48T02Device::Peek(uint32 offset)
{
	if (offset < 2040) {
		return mem[offset];
	} else {
		switch (offset - 2040) {
		 case 0:	// コントロールレジスタ
			return reg.ctrl;
		 case 1:	// 秒 | STOP
			return reg.ex.sec | reg.stop;
		 case 2:	// 分
			return reg.ex.min;
		 case 3:	// 時
			return reg.ex.hour | reg.kickstart;
		 case 4:	// 曜日
			return reg.ex.wday | reg.freqtest;
		 case 5:	// 日
			return reg.ex.mday;
		 case 6:	// 月
			return reg.ex.mon;
		 case 7:	// 年
			return reg.ex.year;
		 default:
			__unreachable();
		}
	}
}

void
MK48T02Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	uint32 addr;
	uint32 v;
	int y = 0;

	screen.Clear();

	// BCD なので %x で表示する
	screen.Print(0, y++,
		"Internal: %04d[%02x]/%02x/%02x(%s) %02x:%02x:%02x.%1x",
		DecodeYear(reg.in.year), reg.in.year,
		reg.in.mon, reg.in.mday, wdays[reg.in.wday],
		reg.in.hour, reg.in.min, reg.in.sec, (uint32)((cnt >> 1) & 0xf));
	screen.Print(0, y++,
		"Register: %04d[%02x]/%02x/%02x(%s) %02x:%02x:%02x",
		DecodeYear(reg.ex.year), reg.ex.year,
		reg.ex.mon, reg.ex.mday, wdays[reg.ex.wday],
		reg.ex.hour, reg.ex.min, reg.ex.sec);
	screen.Print(0, y++, "TimeZone: %s",
		use_localtime ? "Local" : "UTC");
	screen.Print(0, y++, "Epoch   : %4d  (Adjust Leap: %d)",
		year_epoch, adjust_leap);
	y++;

	addr = 0x800;
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (year=%02x)", addr, v, v);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (mon =%02x)", addr, v, v);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (date=%02x)", addr, v, v);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (wday=%02x  FT=%d)",
		addr, v, (v & 0x07), (v & MK48T02Clock::FREQTEST) ? 1 : 0);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (hour=%02x  KS=%d)",
		addr, v, (v & 0x3f), (v & MK48T02Clock::KICKSTART) ? 1 : 0);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (min =%02x)", addr, v, (v & 0x7f));
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (sec =%02x  ST=%d)",
		addr, v, (v & 0x7f), (v & MK48T02Clock::STOP) ? 1 : 0);
	v = Peek(--addr);
	screen.Print(0, y++, "$%3x: $%02x (W=%d R=%d)",
		addr, v,
		(v & MK48T02Clock::WRITE) ? 1 : 0,
		(v & MK48T02Clock::READ)  ? 1 : 0);
}

// コントロールレジスタへの書き込み
void
MK48T02Device::WriteCtrl(uint32 data)
{
	bool write_fall = bit_falling(reg.ctrl, data, MK48T02Clock::WRITE);

	reg.ctrl = data;

	// ホールドしているかどうかは状態なのでエッジではない
	hold = (reg.ctrl & (MK48T02Clock::READ | MK48T02Clock::WRITE));

	if (write_fall) {
		// WRITE ビットの立ち下がりエッジで
		// 外部レジスタから内部カウンタへセット
		reg.in = reg.ex;
		// 分周段は規定されていないが意味的にクリアするほうが良いだろう
		cnt = 0;
	}
}

uint
MK48T02Device::GetSec() const
{
	return BCD2num(reg.in.sec);
}

void
MK48T02Device::SetSec(uint v)
{
	reg.in.sec = num2BCD(v);
}

uint
MK48T02Device::GetMin() const
{
	return BCD2num(reg.in.min);
}

void
MK48T02Device::SetMin(uint v)
{
	reg.in.min = num2BCD(v);
}

uint
MK48T02Device::GetHour() const
{
	return BCD2num(reg.in.hour);
}

void
MK48T02Device::SetHour(uint v)
{
	reg.in.hour = num2BCD(v);
}

uint
MK48T02Device::GetWday() const
{
	// データシートの正規の値範囲は 1..7 だが 0 も問題なく保持できて、
	// NetBSD/luna68k, OpenBSD/luna88k は日付を日曜に変更する際には 0 を
	// レジスタに書き込んでいる。そのためレジスタ値は 0 と 7 を日曜、
	// 1を月曜…という運用ということにする。
	// 戻り値は 0..6 (0が日曜)。

	int v = reg.in.wday;
	if (v == 7) {
		v = 0;
	}
	return v;
}

void
MK48T02Device::SetWday(uint v)
{
	// 引数は 0..6 (0が日曜)、MK48T02 の正規の範囲は 1..7 で曜日との対応は
	// 運用依存のようだ。NetBSD/luna68k, OpenBSD/luna88k は 0 を日曜、1 を
	// 月曜…と扱っているが、MK48T02 の曜日レジスタは 7 の翌日 1 に繰り上がる
	// 動作なので、ここでもレジスタには 1..7 (1が月曜、7が日曜) をセットする。
	if (v == 0)
		v = 7;
	reg.in.wday = v;
}

uint
MK48T02Device::GetMday() const
{
	return BCD2num(reg.in.mday);
}

void
MK48T02Device::SetMday(uint v)
{
	reg.in.mday = num2BCD(v);
}

uint
MK48T02Device::GetMon() const
{
	return BCD2num(reg.in.mon);
}

void
MK48T02Device::SetMon(uint v)
{
	reg.in.mon = num2BCD(v);
}

uint
MK48T02Device::DecodeYear(uint bcd_year) const
{
	uint y = BCD2num(bcd_year);
	y += year_epoch;

	// 1970年より前を 2000 年代とみなすのはシステム依存だが
	// UNIX 系はみんなこう(?)
	// XXX ここの 1970 と RTC エポックを一致させれば端折れるじゃん！と
	//     思いついたのかしら?
	if (y < 1970) {
		y += 100;
	}
	return y;
}

uint
MK48T02Device::GetYear() const
{
	// see DecodeYear()
	return DecodeYear(reg.in.year);
}

void
MK48T02Device::SetYear(uint v)
{
	// see DecodeYear()
	v -= year_epoch;
	reg.in.year = num2BCD(v % 100);
}

uint
MK48T02Device::GetLeap() const
{
	// 補正項込みで計算
	return (BCD2num(reg.in.year) + adjust_leap) & 3;
}

void
MK48T02Device::SetLeap(uint v)
{
	// nop
	// MK48T02 ではうるう年カウンタはなく、西暦年でうるう年を判断している。
}

// 1Hz 仮想 RTC からのコール
// 分周段 0/32 秒でコールされる。
void
MK48T02Device::Tick1Hz()
{
	if (reg.stop) {
		// 実機は電力消費を抑えるため発振器ごと止めるが、
		// 分周段が規定されていないので秒カウントを止めれば十分なはず
		//nop
	} else {
		CountUpSec();
	}

	if (hold == false) {
		// 毎秒コピーする (Datasheet p2)
		reg.ex = reg.in;
	}
}

// 指定したアドレスからのゼロ終端文字列を返す
// (内蔵 ROM からのアクセス用)
std::string
MK48T02Device::PeekString(uint32 addr) const
{
	std::string rv = "";

	addr &= 0x7ff;
	for (; mem[addr]; addr++) {
		rv += mem[addr];
	}
	return rv;
}

// 指定したアドレスに文字列を書き込み、
// 16 バイトに満たないところはゼロでパディングする。
// 書き込めれば true を、書き込めなければ false を返す。
// (内蔵 ROM からの書き込み用)
bool
MK48T02Device::WriteString(uint32 addr, const std::string& data)
{
	int i = 0;

	// ここの文字列は16バイト未満なら一見ゼロ終端のようだが、
	// 16バイトならゼロ終端せず16文字とするフォーマット。

	if (data.length() > 16) {
		return false;
	}

	addr &= 0x7ff;
	for (; i < data.length(); i++) {
		mem[addr + i] = data[i];
	}
	for (; i < 16; i++) {
		mem[addr + i] = 0;
	}
	return true;
}

// 全域をゼロクリアする。
// クリアするだけなので、その後何か書いたりチェックサムを計算したりする
// 必要はある。
// (内蔵 ROM からの書き込み用)
void
MK48T02Device::ClearAll()
{
	for (int i = 0, sz = file.GetMemSize(); i < sz; i++) {
		mem[i] = 0;
	}
}
