//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// MPU (m680x0)

#include "mpu680x0.h"
#include "config.h"
#include "m68030bus.h"
#include "m68030excep.h"
#include "mainapp.h"
#include "uimessage.h"
#include "vm.h"
#include <math.h>

// コンストラクタ
MPU680x0Device::MPU680x0Device()
	: inherited("MPU(68030)")
{
	cpu = m68030_init();

	// イベントコールバック設定
	event.SetName("MPU");
	event.func = (DeviceCallback_t)&MPU680x0Device::Callback;

	// レジスタモニター
	reg_monitor.func = (MonitorCallback_t)&MPU680x0Device::MonitorUpdateReg;
	reg_monitor.SetSize(79, 16);
	reg_monitor.Regist(ID_MONITOR_MPUREG);

	// ATC モニター
	atc_monitor.func = (MonitorCallback_t)&MPU680x0Device::MonitorUpdateATC;
#if !defined(ATC_SINGLE)
	atc_monitor.SetSize(131, m68030ATCTable::LINES + 3);
#else
	atc_monitor.SetSize(31, 24);
#endif
	atc_monitor.Regist(ID_MONITOR_MPUATC);
}

// デストラクタ
MPU680x0Device::~MPU680x0Device()
{
}

bool
MPU680x0Device::Init()
{
	// 親クラス
	if (inherited::Init() == false) {
		return false;
	}
	// MPU クロックは親 Init() で読み込んで、ここで cpu にセットする
	m68030_set_clockspeed(cpu, clock_khz);
	// 割り込み時のディレイ
	intdelay = cpu->Cycle2Vtime(4);

	// FPU (6888x)
	if (gMainApp.GetVMType() == VMTYPE_LUNA1) {
		// LUNA は 68881 固定 (設定は見ない)
		cpu->has_fpu = 1;
	} else {
		// X68k は設定による
		const ConfigItem& item = gConfig->Find("mpu-has-fpu");
		cpu->has_fpu = item.AsInt();
		if (cpu->has_fpu < 0 || cpu->has_fpu > 2) {
			item.Err();
			return false;
		}
	}

	cpu->halt_callback = &MPU680x0Device::HaltCallback;
	cpu->halt_arg = NULL;

	return true;
}

bool
MPU680x0Device::PowerOn()
{
	m68030_power_on(cpu);
	return true;
}

void
MPU680x0Device::ResetHard()
{
	// リセット例外
	m68030_request_reset_exception(cpu);
}

// イベントコールバック
// ev.code は
// 0x0 〜 0x7: 新しい割り込みレベルを通知 (0-7)
void
MPU680x0Device::Callback(Event& ev)
{
	int level = ev.code;
	m68030_interrupt(cpu, level);
}

// MPU を vtime [nsec] 分実行する。
// 戻り値は outer フラグ。
uint32
MPU680x0Device::Run(uint32 vtime)
{
	return m68030_run(cpu, vtime);
}

// 現在の仮想時刻 [nsec] を取得する
uint64
MPU680x0Device::GetVirtTime() const
{
	return m68030_get_vtime(cpu);
}

// 割り込みレベルが変わったことを MPU に通知。
void
MPU680x0Device::Interrupt(int level)
{
	assertmsg(0 <= level && level <= 7, "level=%d", level);

	// 外部デバイスが 68030 の IPL0-2 をアサートしてから、68030 がそれを
	// 認識(確定)するのに 2クロックかかる。ここでレベルを比較して割り込みを
	// 上げる場合は /IPEND をアサートする。
	// そこからさらに2クロック後以降の命令境界で割り込みを受け付ける(?)
	// らしい?
	// この 4クロック待つところはイベントスケジューラがある VM 側で担当する。
	// 本来、レベル比較はここから2クロック後に行われるはずだが、ちょっと省略。
	// 今はコア側でペンディング処理をしているのでこっち側は起きたことを全部
	// 通知する。
	// 割り込みレベルが 0 に下がる時にはこのディレイは不要のはず。
	if (level != 0) {
		event.time = intdelay;
	} else {
		event.time = 0;
	}
	event.code = level;
	event.Start();
}

// A-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetALineCallback(bool (*callback)(m68kcpu *, void*), void *arg)
{
	cpu->aline_callback = callback;
	cpu->aline_arg = arg;
}

// F-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetFLineCallback(bool (*callback)(m68kcpu *, void*), void *arg)
{
	cpu->fline_callback = callback;
	cpu->fline_arg = arg;
}

// MPU ホールト時の MPU からのコールバック関数
/*static*/ void
MPU680x0Device::HaltCallback(void *arg)
{
	UIMessage::Post(UIMessage::HALT);
}

// MPU からのアクセスをエミュレート
uint64
MPU680x0Device::Read8(uint32 addr)
{
	try {
		return m68030_read_8(cpu, addr);
	} catch (...) {
		return (uint64)-1;
	}
}

uint64
MPU680x0Device::Read16(uint32 addr)
{
	try {
		return m68030_read_16(cpu, addr);
	} catch (...) {
		return (uint64)-1;
	}
}

uint64
MPU680x0Device::Read32(uint32 addr)
{
	try {
		return m68030_read_32(cpu, addr);
	} catch (...) {
		return (uint64)-1;
	}
}

uint64
MPU680x0Device::Write8(uint32 addr, uint32 data)
{
	try {
		m68030_write_8(cpu, addr, data);
		return 0;
	} catch (...) {
		return (uint64)-1;
	}
}

uint64
MPU680x0Device::Write16(uint32 addr, uint32 data)
{
	try {
		m68030_write_16(cpu, addr, data);
		return 0;
	} catch (...) {
		return (uint64)-1;
	}
}

uint64
MPU680x0Device::Write32(uint32 addr, uint32 data)
{
	try {
		m68030_write_32(cpu, addr, data);
		return 0;
	} catch (...) {
		return (uint64)-1;
	}
}

// モニター更新 (レジスタウィンドウ)
void
MPU680x0Device::MonitorUpdateReg(Monitor *, TextScreen& screen)
{
	m68kreg reg;
	uint32 ppc;
	int x;
	int y;

	screen.Clear();

	// ローカルにコピー
	// ただし PPC だけ reg 構造体ではなく cpu クラス側にある
	reg = cpu->reg;
	ppc = cpu->ppc;

	// データレジスタ、アドレスレジスタ
	for (int i = 0; i < 4; i++) {
		screen.Print(0, i, "D%d:%08x D%d:%08x  A%d:%08x A%d:%08x",
			i + 0, reg.da[i],
			i + 4, reg.da[i + 4],
			i + 0, reg.da[i + 8],
			i + 4, reg.da[i + 12]);
	}

	// 5列目
	x = 50;
	// SR
	screen.Print(x, 0, "SR:%04x(%c%c%c%c%c)",
		(reg.sr_h() | reg.ccr.Get()),
		(reg.ccr.IsX()) ? 'X' : '-',
		(reg.ccr.IsN()) ? 'N' : '-',
		(reg.ccr.IsZ()) ? 'Z' : '-',
		(reg.ccr.IsV()) ? 'V' : '-',
		(reg.ccr.IsC()) ? 'C' : '-');

	// VBR
	screen.Print(x, 1, "VBR:%08x", reg.vbr);

	// *SP
	uint ms = (reg.s ? 2 : 0) | (reg.m ? 1 : 0); // T1 T0 S M
	if (ms == 3) {
		// Supervisor (Master) mode (A7=MSP, USP/ISP)
		screen.Print(x, 2, TA::Disable, "USP:%08x", reg.usp);
		screen.Print(x, 3, TA::Disable, "ISP:%08x", reg.isp);
	} else if (ms == 2) {
		// Supervisor (Interrupt) mode (A7=ISP, USP/MSP)
		screen.Print(x, 2, TA::Disable, "USP:%08x", reg.usp);
		screen.Print(x, 3, TA::Disable, "MSP:%08x", reg.msp);
	} else {
		// User mode (A7=USP, ISP/MSP)
		screen.Print(x, 2, TA::Disable, "ISP:%08x", reg.isp);
		screen.Print(x, 3, TA::Disable, "MSP:%08x", reg.msp);
	}

	// 6列目
	// PC:  01234567
	// SFC:1   DFC:1
	// CACR:01234567
	// CAAR:01234567
	x = 66;
	screen.Print(x, 0, "PC:  %08x", ppc);
	screen.Print(x, 1, "SFC:%d  DFC:%d", reg.sfc, reg.dfc);
	screen.Print(x, 2, "CACR:%08x", reg.cacr);
	screen.Print(x, 3, "CAAR:%08x", reg.caar);

	// XXX どこに置くのがいいか
	if (reg.state == 0) {
		screen.Puts(25, 4, "State: Running");
	} else if (reg.state == CPU_REQ_STOP) {
		screen.Puts(25, 4, "State: STOP instruction");
	} else if (reg.state == CPU_REQ_HALT) {
		screen.Puts(25, 4, TA::On, "State: Double Bus Fault");
	}

	// MMU
	x = 0;
	y = 4;
	screen.Puts(x, y++, "<MMU>");
	screen.Print(x, y,     "SRP:%08x_%08x", reg.srp.h, reg.srp.l);
	screen.Print(x, y + 1, "CRP:%08x_%08x", reg.crp.h, reg.crp.l);
	x += 23;
	for (int i = 0; i < 2; i++) {
		uint32 tt = reg.tt[i];
		screen.Print(x, y + i, "TT%d:%08x(%c%c%c)",
			i, tt,
			(tt & m68030TT::E)   ? 'E' : '-',
			(tt & m68030TT::CI)  ? 'C' : '-',
			(tt & m68030TT::RWM) ? '-' : ((tt & m68030TT::RW)  ? 'R' : 'W'));
	}
	x += 19;
	screen.Print(x, y, "TC:%08x(%c%c%c)",
		reg.tc,
		(reg.tc & m68030TC::TC_E)   ? 'E' : '-',
		(reg.tc & m68030TC::TC_SRE) ? 'S' : '-',
		(reg.tc & m68030TC::TC_FCL) ? 'F' : '-');
	screen.Print(x, y + 1, "MMUSR: %04x(%c%c%c%c%c%c%c N=%d)",
		reg.mmusr,
		(reg.mmusr & m68030MMUSR::B) ? 'B' : '-',
		(reg.mmusr & m68030MMUSR::L) ? 'L' : '-',
		(reg.mmusr & m68030MMUSR::S) ? 'S' : '-',
		(reg.mmusr & m68030MMUSR::W) ? 'W' : '-',
		(reg.mmusr & m68030MMUSR::I) ? 'I' : '-',
		(reg.mmusr & m68030MMUSR::M) ? 'M' : '-',
		(reg.mmusr & m68030MMUSR::T) ? 'T' : '-',
		(reg.mmusr & m68030MMUSR::N));

	// FPU
	x = 0;
	y = 7;
	screen.Puts(x, y++, "<FPU>");
	for (int i = 0; i < 8; i++) {
		screen.Print(x, y + i, "FP%d:%04x_%08x_%08x (%-20s)",
			i,
			reg.fpframe.fpf_regs[i * 3 + 0] >> 16,
			reg.fpframe.fpf_regs[i * 3 + 1],
			reg.fpframe.fpf_regs[i * 3 + 2],
			"not yet");
	}

	x = 51;

	// FPCR
	static const char * const rpstr[] = {
		".EXT",
		".SGL",
		".DBL",
		".???",
	};
	static const char * const rmstr[] = {
		"Near",
		"Zero",
		"Minus",
		"Plus",
	};
	uint32 fpcr = reg.fpframe.fpf_fpcr;
	screen.Print(x, y++, "FPCR:%04x", fpcr);
	screen.Puts(x +  2, y, TA::OnOff(fpcr & 0x8000), "BS");
	screen.Puts(x +  5, y, TA::OnOff(fpcr & 0x4000), "SN");
	screen.Puts(x +  8, y, TA::OnOff(fpcr & 0x2000), "OP");
	screen.Puts(x + 11, y, TA::OnOff(fpcr & 0x1000), "OV");
	screen.Puts(x + 14, y, TA::OnOff(fpcr & 0x0800), "UF");
	screen.Puts(x + 17, y, TA::OnOff(fpcr & 0x0400), "DZ");
	screen.Puts(x + 20, y, TA::OnOff(fpcr & 0x0200), "I2");
	screen.Puts(x + 23, y, TA::OnOff(fpcr & 0x0100), "I1");
	y++;
	screen.Print(x + 2, y++, "RP=%s RM=%s",
		rpstr[(fpcr >> 6) & 3],
		rmstr[(fpcr >> 4) & 3]);

	// FPSR
	uint32 fpsr = reg.fpframe.fpf_fpsr;
	uint32 cc = fpsr >> 24;
	screen.Print(x, y++, "FPSR:%08x", fpsr);
	screen.Puts(x +  2, y, TA::OnOff(cc & 0x08), "N");
	screen.Puts(x +  4, y, TA::OnOff(cc & 0x04), "Z");
	screen.Puts(x +  6, y, TA::OnOff(cc & 0x02), "Inf");
	screen.Puts(x + 10, y, TA::OnOff(cc & 0x01), "NAN");
	screen.Print(x + 14, y, "Q=$%02x", (fpsr >> 16) & 0xff);
	y++;
	screen.Puts(x +  2, y, TA::OnOff(fpsr & 0x8000), "BS");
	screen.Puts(x +  5, y, TA::OnOff(fpsr & 0x4000), "SN");
	screen.Puts(x +  8, y, TA::OnOff(fpsr & 0x2000), "OP");
	screen.Puts(x + 11, y, TA::OnOff(fpsr & 0x1000), "OV");
	screen.Puts(x + 14, y, TA::OnOff(fpsr & 0x0800), "UF");
	screen.Puts(x + 17, y, TA::OnOff(fpsr & 0x0400), "DZ");
	screen.Puts(x + 20, y, TA::OnOff(fpsr & 0x0200), "I2");
	screen.Puts(x + 23, y, TA::OnOff(fpsr & 0x0100), "I1");
	y++;
	screen.Puts(x +  2, y, TA::OnOff(fpsr & 0x80), "IOP");
	screen.Puts(x +  6, y, TA::OnOff(fpsr & 0x80), "OVFL");
	screen.Puts(x + 11, y, TA::OnOff(fpsr & 0x80), "UNFL");
	screen.Puts(x + 16, y, TA::OnOff(fpsr & 0x80), "DZ");
	screen.Puts(x + 19, y, TA::OnOff(fpsr & 0x80), "INEX");
	y++;

	// FPIAR
	screen.Print(x, y++, "FPIAR:%08x", reg.fpframe.fpf_fpiar);
}

// モニター更新 (ATC)
void
MPU680x0Device::MonitorUpdateATC(Monitor *, TextScreen& screen)
{
	m68030ATCTable *table;
#if !defined(ATC_SINGLE)
	struct {
		int fc;
		const char *name;
	} fclist[] = {
		{ 1,	"User/Data" },
		{ 2,	"User/Program" },
		{ 5,	"Super/Data" },
		{ 6,	"Super/Program" },
	};

	screen.Clear();

	// 行数は左端にだけ
	for (int i = 0; i < m68030ATCTable::LINES; i++) {
		screen.Print(0, i + 2, "%02d:", i);
	}

	for (int t = 0; t < countof(fclist); t++) {
		int x = t * 32 + 4;
		int fc = fclist[t].fc;
		const char *name = fclist[t].name;
		double r;
		int i;

		table = &cpu->atc.tables[fc];
		screen.Print(x, 0, "FC=%d %s", fc, name);
		screen.Puts(x, 1, "LAddr    PAddr   Flag Hit%  Age");
		// 先に統計用の母数を計算
		uint64 total = 0;
		for (i = 0; i < countof(table->hit); i++) {
			total += table->hit[i];
		}
		total += table->hit_head;

		// エントリ表示
		for (i = 0; i < countof(table->line); i++) {
			m68030ATCLine& a = table->line[i];
			TA attr;
			if (a.IsInvalid()) {
				attr = TA::Disable;
			} else if (table->head == &a) {
				attr = TA::Em;
			} else {
				attr = TA::Normal;
			}

			screen.Print(x, i + 2, attr, "%08x %08x %c%c%c",
				a.GetLAddr(),
				a.GetPAddr(),
				a.IsBusError() ? 'B' : '-',
				a.IsWProtect() ? 'P' : '-',
				a.IsModified() ? 'M' : '-');
		}
		// ヘッドのヒット率
		screen.Puts(x + 0, i + 2, "head");
		r = (double)table->hit_head * 100 / total;
		if (!isnan(r)) {
			screen.Print(x + 5, i + 2, "%4.1f%%", r);
		}
		// 配列のヒット率とミス率
		screen.Puts(x + 17, i + 2, "miss");
		for (i = 0; i < countof(table->hit); i++) {
			r = (double)(table->hit[i] * 100) / total;
			if (!isnan(r)) {
				screen.Print(x + 22, i + 2, "%4.1f%%", r);
			}
		}
		// カウンタ
		// 分かりやすさのため最大値からの差(距離)を表示
		for (i = 0; i < countof(table->line); i++) {
			m68030ATCLine& a = table->line[i];
			if (!a.IsInvalid()) {
				screen.Print(x + 27, i + 2, "%4d", table->head->age - a.age);
			}
		}
	}
#else
	m68030ATCLine *a;
	int i;

	table = &cpu->atc.tables[0];
	screen.Clear();

	screen.Puts(0, 0, "No. FC    LAddr    PAddr   Flag");
	for (a = table->head, i = 0; a; a = a->next, i++) {
		screen.Print(0, i + 1, "%02d:", i);
		if (!a->IsInvalid()) {
			char dp;
			switch (a->fc & 3) {
			 case 1:	dp = 'D';	break;
			 case 2:	dp = 'P';	break;
			 default:	dp = '?';	break;
			}
			screen.Print(4, i + 1, "%d(%c%c):%08x %08x %c%c%c",
				a->fc,
				(a->fc & 4) ? 'S' : 'U',
				dp,
				a->GetLAddr(), a->GetPAddr(),
				a->IsBusError() ? 'B' : '-',
				a->IsWProtect() ? 'W' : '-',
				a->IsModified() ? 'M' : '-');
		}
	}
#endif
}

// CPU 側からのコールバック。RESET 命令ハンドラ
void
m68030_reset_inst(m68kcpu *cpu)
{
	gVM->ResetByMPU();
}
