//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// MPU (m88xx0)

#include "mpu88xx0.h"
#include "config.h"
#include "sysctlr.h"

// コンストラクタ
MPU88xx0Device::MPU88xx0Device()
	: inherited("MPU(88xx0)")
{
	cpu.reset(new m88kcpu());

	// LUNA88K では CMMU の ID は次のように割り振ってあるようだ。
	// OpenBSD の sys/arch/luna68k/include/board.h より。
	//
	//       Inst Data
	// CPU#0 ID=7 ID=6
	// CPU#1 ID=5 ID=4
	// CPU#2 ID=3 ID=2
	// CPU#3 ID=1 ID=0
	//
	// 厳密にはこれは LUNA88K のハードウェアがどういうコンフィグレーションで
	// m88200 を使うかなので、ここではなくて vm_luna あたりのほうがいいかも
	// しれんけど、あまり遠いのも面倒なので、とりあえずこの辺に。
	// あとマルチプロセッサになったらまたその時考える。
	cpu->cmmu[0].SetID(7);
	cpu->cmmu[1].SetID(6);

	// レジスタモニター
	monitor.func = (MonitorCallback_t)&MPU88xx0Device::MonitorUpdate;
	monitor.SetSize(79, 18);
	monitor.Regist(ID_MONITOR_MPUREG);
}

// デストラクタ
MPU88xx0Device::~MPU88xx0Device()
{
}

bool
MPU88xx0Device::Init()
{
	// 親クラス
	if (inherited::Init() == false) {
		return false;
	}
	// MPU クロックは親 Init() で読み込んで、ここで cpu にセットする
	cpu->SetClockSpeed(clock_khz);

	const ConfigItem& item = gConfig->Find("mpu-pseudo-stop");
	cpu->pseudo_stop_enable = (bool)item.AsInt();

	return true;
}

bool
MPU88xx0Device::PowerOn()
{
	cpu->PowerOn();
	return true;
}

void
MPU88xx0Device::ResetHard()
{
	cpu->RequestReset();
}

// アクセス中の論理アドレスを取得
uint32
MPU88xx0Device::GetLaddr() const
{
	m88200 *cmmu = m88200::GetBusMaster();
	assert(cmmu);
	return cmmu->GetLaddr();
}

// アクセス中の物理アドレスを取得
uint32
MPU88xx0Device::GetPaddr() const
{
	m88200 *cmmu = m88200::GetBusMaster();
	assert(cmmu);
	return cmmu->GetPaddr();
}

// アクセスウェイトを加算
void
MPU88xx0Device::AddCycle(int32 cycle)
{
	// 今アクセスしてきている(= バスマスターの) CMMU に対して加算する
	m88200 *cmmu = m88200::GetBusMaster();
	assert(cmmu);
	cmmu->AddCycle(cycle);
}

// Interrupt
void
MPU88xx0Device::Interrupt(int level)
{
	cpu->Interrupt(level);
}

// DOS call エミュレーションのコールバックを指定
void
MPU88xx0Device::SetFLineCallback(bool (*callback)(m88kcpu *, void *), void *arg)
{
	cpu->SetFLineCallback(callback, arg);
}

// MPU からのアクセスをエミュレート
uint64
MPU88xx0Device::Read8(uint32 addr)
{
	return cpu->cmmu[1].load_8(addr);
}

uint64
MPU88xx0Device::Read16(uint32 addr)
{
	if ((addr & 1) == 0) {
		return cpu->cmmu[1].load_16(addr);
	} else {
		uint64 h, l;
		h = Read8(addr);
		if ((int64)h < 0) {
			return h;
		}
		l = Read8(addr + 1);
		if ((int64)l < 0) {
			return l;
		}
		return (h << 8) | l;
	}
}

uint64
MPU88xx0Device::Read32(uint32 addr)
{
	if ((addr & 3) == 0) {
		return cpu->cmmu[1].load_32(addr);
	} else if ((addr & 1) == 0) {
		uint64 h, l;
		h = Read16(addr);
		if ((int64)h < 0) {
			return h;
		}
		l = Read16(addr + 2);
		if ((int64)l < 0) {
			return l;
		}
		return (h << 16) | l;
	} else {
		uint64 h, m, l;
		h = Read8(addr);
		if ((int64)h < 0) {
			return h;
		}
		m = Read16(addr + 1);
		if ((int64)m < 0) {
			return m;
		}
		l = Read8(addr + 3);
		if ((int64)l < 0) {
			return l;
		}
		return (h << 24) | (m << 8) | l;
	}
}

uint64
MPU88xx0Device::Write8(uint32 addr, uint32 data)
{
	return cpu->cmmu[1].store_8(addr, data & 0xff);
}

uint64
MPU88xx0Device::Write16(uint32 addr, uint32 data)
{
	if ((addr & 1) == 0) {
		return cpu->cmmu[1].store_16(addr, data & 0xffff);
	} else {
		uint64 r;
		r = Write8(addr, data >> 8);
		if ((int64)r < 0) {
			return r;
		}
		return Write8(addr + 1, data);
	}
}

uint64
MPU88xx0Device::Write32(uint32 addr, uint32 data)
{
	if ((addr & 3) == 0) {
		return cpu->cmmu[1].store_32(addr, data);
	} else if ((addr & 1) == 0) {
		uint64 r;
		r = Write16(addr, data >> 16);
		if ((int64)r < 0) {
			return r;
		}
		return Write16(addr + 2, data);
	} else {
		uint64 r;
		r = Write8(addr, data >> 24);
		if ((int64)r < 0) {
			return r;
		}
		r = Write16(addr + 1, data >> 8);
		if ((int64)r < 0) {
			return r;
		}
		return Write8(addr + 3, data);
	}
}

// モニター更新
void
MPU88xx0Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	m88100reg reg;
	int x;
	int y;

	screen.Clear();

	// ローカルにコピー
	reg = *cpu;

	//
	for (int i = 0; i < countof(reg.r); i++) {
		screen.Print((i / 8) * 14, i % 8, "r%-2d:%08x", i, reg.r[i]);
	}

	// 制御レジスタ
	x = 0;
	y = 9;
	screen.Print(x, y++,  "pid (cr0):%08x(Arch=$%02x Ver=$%02x MC=%s)",
		reg.pid,
		(reg.pid >> 8) & 0xff,
		(reg.pid >> 1) & 0x7f,
		(reg.pid & 1) ? "Master" : "Checker");
	screen.Print(x, y, "psr (cr1):%08x(", reg.psr);
	screen.Puts(x + 19, y, TA::OnOff(reg.psr  & m88100reg::PSR_SUPER), "S");
	screen.Puts(x + 21, y,
		((reg.psr & m88100reg::PSR_BO_LE) ? "BO=LE" : "BO=BE"));
	screen.Puts(x + 27, y, TA::OnOff(reg.psr  & m88100reg::PSR_SER),  "SER");
	screen.Puts(x + 31, y, TA::OnOff(reg.psr  & m88100reg::PSR_C),    "Cy");
	screen.Puts(x + 34, y, TA::OnOff(reg.psr  & m88100reg::PSR_SFD1), "SFD1");
	screen.Puts(x + 39, y, TA::OnOff(reg.psr  & m88100reg::PSR_MXM),  "MXM");
	screen.Puts(x + 43, y, TA::OnOff(reg.psr  & m88100reg::PSR_IND),  "IND");
	screen.Puts(x + 47, y, TA::OnOff(reg.psr  & m88100reg::PSR_SFRZ), "SFRZ");
	screen.Puts(x + 51, y, ")");
	y++;
	screen.Print(x, y, "epsr(cr2):%08x(", reg.epsr);
	screen.Puts(x + 19, y, TA::OnOff(reg.epsr & m88100reg::PSR_SUPER), "S");
	screen.Puts(x + 21, y,
		((reg.epsr& m88100reg::PSR_BO_LE) ? "BO=LE" : "BO=BE"));
	screen.Puts(x + 27, y, TA::OnOff(reg.epsr & m88100reg::PSR_SER),  "SER");
	screen.Puts(x + 31, y, TA::OnOff(reg.epsr & m88100reg::PSR_C),    "Cy");
	screen.Puts(x + 34, y, TA::OnOff(reg.epsr & m88100reg::PSR_SFD1), "SFD1");
	screen.Puts(x + 39, y, TA::OnOff(reg.epsr & m88100reg::PSR_MXM),  "MXM");
	screen.Puts(x + 43, y, TA::OnOff(reg.epsr & m88100reg::PSR_IND),  "IND");
	screen.Puts(x + 47, y, TA::OnOff(reg.epsr & m88100reg::PSR_SFRZ), "SFRZ");
	screen.Puts(x + 51, y, ")");

	// 制御レジスタ(左中; *IP)
	x = 0;
	y = 12;
	screen.Print(x, y + 0, "xip:%08x  opX:%08x(%c%c)",
		reg.xip, (uint32)reg.opX,
		cpu->OpIsBusErr(reg.opX) ? 'B' : '-',
		cpu->OpIsDelay(reg.opX)  ? 'D' : '-');
	screen.Print(x, y + 1, "nip:%08x  opF:%08x(%c%c)",
		reg.nip, (uint32)reg.opF,
		cpu->OpIsBusErr(reg.opF) ? 'B' : '-',
		cpu->OpIsDelay(reg.opF)  ? 'D' : '-');
	screen.Print(x, y + 2, "fip:%08x", reg.fip);

	// 制御レジスタ (S*IP)
	x = 32;
	for (int i = 0; i < 3; i++) {
		screen.Print(x, y + i, "%s(cr%d):%08x:%c%c",
			m88100reg::sipname[i], 4 + i,
			(reg.cr[4 + i] & m88100reg::SIP_MASK),
			(reg.cr[4 + i] & m88100reg::SIP_V) ? 'V' : '-',
			(reg.cr[4 + i] & m88100reg::SIP_E) ? 'E' : '-');
	}

	// 制御レジスタ(右中)
	x = 55;
	y = 9;
	for (int i = 0; i < 4; i++) {
		int rn = 17 + i;
		screen.Print(x, y++, "sr%d(cr%d):%08x", i, rn, reg.cr[rn]);
	}
	screen.Print(x, y++, "ssbr(cr3):%08x", reg.ssbr);
	screen.Print(x, y++, "vbr (cr7):%08x", reg.vbr);

	// MMU レジスタ(下段)
	y = 15;
	for (int i = 0; i <= 2; i++) {
		int dt =  8 + i * 3;
		int dd =  9 + i * 3;
		int da = 10 + i * 3;

		screen.Print(0, y, "dmt%d(cr%-2d):%04x",
			i, dt, (reg.cr[dt] & 0xffff));
		screen.Puts(15, y, "(b,s,d,l,rx, s,en  ,w,v)");
		screen.Puts(16, y, (reg.cr[dt] & m88100reg::DM_BO)     ? "B" : "-");
		screen.Puts(18, y, (reg.cr[dt] & m88100reg::DM_DAS)    ? "S" : "U");
		screen.Puts(20, y, (reg.cr[dt] & m88100reg::DM_DOUB1)  ? "D" : "-");
		screen.Puts(22, y, (reg.cr[dt] & m88100reg::DM_LOCK)   ? "L" : "-");
		// カンマが自然に見えるようにここだけカンマも含めて前詰めで出力
		screen.Print(25, y, "%d,",
			(reg.cr[dt] & m88100reg::DM_DREG_MASK) >> 7);
		screen.Puts(28, y, (reg.cr[dt] & m88100reg::DM_SIGNED) ? "S" : "-");
		uint32 en = (reg.cr[dt] & m88100reg::DM_EN_MASK) >> 2;
		screen.Puts(30, y, m88100reg::dmt_en_str[en]);
		screen.Puts(35, y, (reg.cr[dt] & m88100reg::DM_WRITE)  ? "W" : "-");
		screen.Puts(37, y, (reg.cr[dt] & m88100reg::DM_VALID)  ? "V" : "-");

		screen.Print(40, y, "dmd%d(cr%-2d):%08x", i, dd, reg.cr[dd]);
		screen.Print(60, y, "dma%d(cr%2d):%08x",  i, da, reg.cr[da]);
		y++;
	}
}
