//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "pio.h"
#include "bitops.h"
#include "config.h"
#include "lcd.h"
#include "mainapp.h"
#include "scheduler.h"

//           IODevice
//               |
//               v
//          i8255Device  (8255 としての共通部分)
//               |
//    +----------+----------+
//    v          v          v
// PPIDevice  PIO0Device PIO1Device
// (X68030)   (LUNA)     (LUNA)

std::unique_ptr<PPIDevice> gPPI;
std::unique_ptr<PIO0Device> gPIO0;
std::unique_ptr<PIO1Device> gPIO1;

//
// i8255
//

// コンストラクタ
i8255Device::i8255Device(const std::string& objname_)
	: inherited(objname_)
{
}

// デストラクタ
i8255Device::~i8255Device()
{
}

// リセット
void
i8255Device::ResetHard()
{
	// clears the control register
	// and place ports A, B, and C in input mode.
	// The input latches in ports A, B, and C are not cleared.

	ctrl = 0b10011011;
}

// コントロールポートへの書き込み。
void
i8255Device::WriteCtrl(uint32 data)
{
	if ((data & 0x80)) {
		// 最上位が立っていればモード設定
		ctrl = data;
		putlog(1,
			"モード設定 groupA=%d groupB=%d portA=%s portB=%s portC=%s/%s",
			GetGroupAMode(), GetGroupBMode(),
			PortADir() == DIR_IN ? "IN" : "OUT",
			PortBDir() == DIR_IN ? "IN" : "OUT",
			PortCHDir() == DIR_IN ? "IN" : "OUT",
			PortCLDir() == DIR_IN ? "IN" : "OUT");

		// モード設定すると PC はゼロクリアされる
		for (int pc = 0; pc < 8; pc++) {
			SetPC(pc, 0);
		}
	} else {
		// portC 出力
		//
		// 0  x  x  x  P  P  P  V
		//             |  |  |  +-- 書き込むビットデータ
		//             +--+--+----- 書き込むビット位置

		int pc = (data >> 1) & 7;
		int val = data & 1;
		SetPC(pc, val);
	}
}

void
i8255Device::SetPC(int pc, int data)
{
	// 継承側で用意すること。ここには来ない
	PANIC("not impl");
}

//
// X68k PPI
//

// コンストラクタ
PPIDevice::PPIDevice()
	: inherited("PPI")
{
	devaddr = baseaddr;
	devlen  = 0x2000;
}

// デストラクタ
PPIDevice::~PPIDevice()
{
}

uint64
PPIDevice::Read(uint32 offset)
{
	gMPU->AddCycle(19);	// InsideOut p.135

	switch (offset) {
	 case 0:
	 case 1:
		return 0xff;
	 case 2:
		putlog(0, "$e9a005 PortC 未実装読み込み");
		return 0xff;
	 case 3:
		PANIC("PPI $e9a007 未サポート読み込み");
		break;
	 default:
		__unreachable();
	}
	PANIC("not impl");
}

uint64
PPIDevice::Write(uint32 offset, uint32 data)
{
	gMPU->AddCycle(19);	// InsideOut p.135

	switch (offset) {
	 case 0:
	 case 1:
		break;
	 case 2:
		putlog(0, "$e9a005 未実装書き込み $%02X (無視)", data);
		break;
	 case 3:
		if (data != 0x92) {
			PANIC("PPI $e9a007 未実装書き込み $%02X", data);
		} else {
			putlog(0, "$e9a007 未実装書き込み $%02x (無視)", data);
		}
		break;
	 default:
		__unreachable();
	}
	return 0;
}

uint64
PPIDevice::Peek(uint32 offset)
{
	// XXX
	return 0xff;
}

//
// LUNA PIO0
//
// 仕様は 0x49000000 から 4バイトだが、
// 0x48000000 から 0x4c000000 の手前まで 4バイト単位でミラーが見える。
//
// LUNA の DIP-SW は #1, #2 にそれぞれ8個、計16個ある。
// DIP-SW #1, #2 が設定ファイルの luna-dipsw{1,2} に対応する。
// luna-dipsw{1,2} はいずれも "0"/"1" を8つ並べた文字列であり、
// 先頭が本体表記でいう 1番、末尾側が本体表記でいう 8番に対応する。
// 値は "0" が down、"1" が up を表す。
//
// PIO0Device 内ではこの設定を読み込んで bool dipsw1[8], dipsw2[8] の配列で
// 値を保持している。dipsw1[0] が本体表記1番、dipsw1[7] が本体表記 8番に対応
// する。値は false が down("0")、true が up("1") に対応する。
//
// PIO から読み出せる値のビットの並び順が機種によって異なる。
// LUNA-1  では 1番が LSB、8番が MSB の並び順、
// LUNA88K では 1番が MSB、8番が LSB の並び順。
// 値は、並び順に関わらずどちらも
// %0 が false = down = "0"、%1 が true = up = "1" である。
//
//    #1 #2 #3 #4 #5 #6 #7 #8
//  +-------------------------+
//  |    [] [] [] [] [] [] [] | 本体外観
//  | []                      |
//  +-------------------------+
//
// luna-dipsw1 = "01111111"    ← 設定ファイル
//                ||||||||     ↓ このクラスのメンバ変数
//                |||||||+- dipsw1[7] = true;
//                ||||||+-- dipsw1[6] = true;
//                |||||+--- dipsw1[5] = true;
//                ||||+---- dipsw1[4] = true;
//                |||+----- dipsw1[3] = true;
//                ||+------ dipsw1[2] = true;
//                |+------- dipsw1[1] = true;
//                +-------- dipsw1[0] = false;
//
//              b7   b6   b5   b4   b3   b2   b1   b0
//            +----+----+----+----+----+----+----+----+
// LUNA-1     | #8 | #7 | #6 | #5 | #4 | #3 | #2 | #1 |
// PortA      |  1 |  1 |  1 |  1 |  1 |  1 |  1 |  0 | %1=UP, %0=DOWN
//            +----+----+----+----+----+----+----+----+
//
//              b7   b6   b5   b4   b3   b2   b1   b0
//            +----+----+----+----+----+----+----+----+
// LUNA88K    | #1 | #2 | #3 | #4 | #5 | #6 | #7 | #8 |
// PortA      |  0 |  1 |  1 |  1 |  1 |  1 |  1 |  1 | %1=UP, %0=DOWN
//            +----+----+----+----+----+----+----+----+
//
// LUNA-1:
// #1-1 up なら UNIX をロードして移行。down なら ROM モニタで起動。
// #1-2 up ならディスプレイコンソール。down ならシリアルコンソール。
// #1-3 up - color display, down - force to have monochrome display
// #1-4 up - no write verification,
//      down - verification on every harddisk write operation
// #1-5 up - OS is UniOS-U (SystemV/COFF),
//      down - OS is UniOS (4.3BSD/a.out OMAGIC)
// #1-6 down - force monochrome display?
// #1-7 up - boot from local device, down - boot from network
// #1-8 up - normal boot, down - start diagnostics と書いてあるが
//      down にすると画面真っ暗で詳細不明。
//
// #2-x すべて up にすると /boot が自動的にカーネルを起動。一つでも down が
//      あれば /boot がプロンプトで停止。これは ROM ではなく NetBSD のブート
//      ローダの話。
//      どこか由来のソースコードらしいがたぶんビット演算を間違えている。
//
// LUNA88K:
// #1-1 up なら ROM モニタで停止、down ならマルチユーザブート。
//      OpenBSD カーネルも up なら UserKernelConfig で止まる模様?。
// #1-2 up なら外付けシリアルコンソール、down なら内蔵ビットマップ。
// #1-3..#1-7 は不明?
// #1-8 down なら自己診断後電源オフするようだ。
// #2-x はユーザ(?)

// コンストラクタ
PIO0Device::PIO0Device()
	: inherited("PIO0")
{
	devaddr = 0x49000000;
	devlen  = 4;

	// 機種固有パラメータ

	// LUNA-1 と LUNA88K で使いやすいように初期値を変えておく。
	switch (gMainApp.GetVMType()) {
	 case VMTYPE_LUNA1:
		gConfig->SetDefault("luna-dipsw1", "11110111");
		msb_first = false;
		break;
	 case VMTYPE_LUNA88K:
		gConfig->SetDefault("luna-dipsw1", "11111111");
		msb_first = true;
		break;
	 default:
		__unreachable();
	}
	// DIP-SW#2 はユーザ定義(?)なので(今の所?)どちらも共通。
	gConfig->SetDefault("luna-dipsw2", "11111111");

	monitor.func = (MonitorCallback_t)&PIO0Device::MonitorUpdate;
	monitor.SetSize(22, 11);
	monitor.Regist(ID_MONITOR_PIO0);
}

// デストラクタ
PIO0Device::~PIO0Device()
{
}

bool
PIO0Device::Init()
{
	// XXX 本当はラッチのタイミングが違う気もするけど、とりあえずね

	const ConfigItem& item1 = gConfig->Find("luna-dipsw1");
	const std::string& str1 = item1.AsString();
	if (str1.size() != 8) {
		item1.Err("must be 8 digits");
		return false;
	}

	const ConfigItem& item2 = gConfig->Find("luna-dipsw2");
	const std::string& str2 = item2.AsString();
	if (str2.size() != 8) {
		item2.Err("must be 8 digits");
		return false;
	}
	for (int i = 0; i < 8; i++) {
		dipsw1[i] = (str1[i] != '0');
		dipsw2[i] = (str2[i] != '0');
	}
	return true;
}

// XXX PowerOn か ResetHard 未調査

uint64
PIO0Device::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:
		// PortA
		putlog(1, "DIP-SW 1 読み込み");
		return GetPA();

	 case 1:
		// PortB
		putlog(1, "DIP-SW 2 読み込み");
		return GetPB();

	 case 2:
		// PortC
		data = GetPC();
		putlog(1, "PC 読み込み -> $%02x", data);
		return data;

	 case 3:
		data = ReadCtrl();
		putlog(1, "コントロール読み込み -> $%02x", data);
		return data;

	 default:
		__unreachable();
	}
}

uint64
PIO0Device::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// PIO0 PortA: DIPSW1 (入力のみ)
	 case 1:	// PIO0 PortB: DIPSW2 (入力のみ)
		return 0;

	 case 2:	// PIO0 PortC: ホスト割り込み制御
		for (int i = 0; i < 8; i++) {
			SetPC(i, data & 1);
			data >>= 1;
		}
		return 0;

	 case 3:	// コントロール
		WriteCtrl(data);
		return 0;

	 default:
		__unreachable();
	}
}

uint64
PIO0Device::Peek(uint32 offset)
{
	switch (offset) {
	 case 0:
		return GetPA();

	 case 1:
		return GetPB();

	 case 2:
		return GetPC();

	 case 3:
		return ReadCtrl();

	 default:
		__unreachable();
	}
}

void
PIO0Device::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;

	screen.Clear();

	y = 0;
	screen.Print(0, y++, "PortA(DIPSW1):%c%c%c%c%c%c%c%c",
		(dipsw1[0] ? '1' : '0'),
		(dipsw1[1] ? '1' : '0'),
		(dipsw1[2] ? '1' : '0'),
		(dipsw1[3] ? '1' : '0'),
		(dipsw1[4] ? '1' : '0'),
		(dipsw1[5] ? '1' : '0'),
		(dipsw1[6] ? '1' : '0'),
		(dipsw1[7] ? '1' : '0'));
	screen.Print(0, y++, "PortB(DIPSW2):%c%c%c%c%c%c%c%c",
		(dipsw2[0] ? '1' : '0'),
		(dipsw2[1] ? '1' : '0'),
		(dipsw2[2] ? '1' : '0'),
		(dipsw2[3] ? '1' : '0'),
		(dipsw2[4] ? '1' : '0'),
		(dipsw2[5] ? '1' : '0'),
		(dipsw2[6] ? '1' : '0'),
		(dipsw2[7] ? '1' : '0'));
	screen.Puts(0, y++, "PortC");
	screen.Puts(1, y++, TA::OnOff(xp_reset), "b7: XP Reset");
	screen.Puts(1, y++, TA::OnOff(parity),   "b6: Parity");
	screen.Puts(1, y++, TA::Disable,         "b5: ---");
	screen.Puts(1, y++, TA::OnOff(int5en),   "b4: Int5 Enable");
	screen.Puts(1, y++, TA::OnOff(int5rq),   "b3: Int5 Request");
	screen.Puts(1, y++, TA::OnOff(int1en),   "b2: Int1 Enable");
	screen.Puts(1, y++, TA::Disable,         "b1: ---");
	screen.Puts(1, y++, TA::OnOff(int1rq),   "b0: Intq Request");
}

// PortA (DIP-SW 1) 読み込み
uint32
PIO0Device::GetPA() const
{
	uint32 data = 0;
	for (int i = 0; i < 8; i++) {
		data |= ((int)dipsw1[i]) << i;
	}
	if (msb_first) {
		data = bitrev(data);
	}
	return data & 0xff;
}

// PortB (DIP-SW 2) 読み込み
uint32
PIO0Device::GetPB() const
{
	uint32 data = 0;
	for (int i = 0; i < 8; i++) {
		data |= ((int)dipsw2[i]) << i;
	}
	if (msb_first) {
		data = bitrev(data);
	}
	return data & 0xff;
}

// PortC 読み込み
uint32
PIO0Device::GetPC() const
{
	uint32 data;

	data = 0x00	// XXX 空きビットは何が読めるか?
		 | (int1rq ? 0x01 : 0)
		 | (int1en ? 0x04 : 0)
		 | (int5rq ? 0x08 : 0)
		 | (int5en ? 0x10 : 0)
		 | (parity ? 0x40 : 0)
		 | (xp_reset?0x80 : 0);
	return data;
}

// PortC 書き込み
void
PIO0Device::SetPC(int pc, int val)
{
	switch (pc) {
	 case 0:
		if ((val ^ int1rq)) {
			int1rq = val;
			putlog(0, "INT1割り込み要求 %d (未実装)", int1rq);
		}
		return;
	 case 1:
		// Not Connected.
		return;
	 case 2:
		if ((val ^ int1en)) {
			int1en = val;
			putlog(0, "INT1割り込み許可 %d (未実装)", int1en);
		}
		return;
	 case 3:
		if ((val ^ int5rq)) {
			int5rq = val;
			putlog(0, "INT5割り込み要求 %d (未実装)", int5rq);
		}
		return;
	 case 4:
		if ((val ^ int5en)) {
			int5en = val;
			putlog(0, "INT5割り込み許可 %d (未実装)", int5en);
		}
		return;
	 case 5:
		// Not Connected.
		return;
	 case 6:
		// パリティ有効/無効
		// 使わないので無視するが読み返しのために覚えておく
		parity = val;
		return;
	 case 7:
		// XP リセット
		if ((val ^ xp_reset)) {
			xp_reset = val;
			putlog(0, "XPリセット %d (未実装)", xp_reset);
		}
		return;
	}
	PANIC("invalid pc=%d", pc);
}


//
// LUNA PIO1
//
// 仕様は 0x4d000000 から 4バイトだが、
// 0x4c000000 から 0x50000000 の手前まで 4バイト単位でミラーが見える。

// コンストラクタ
PIO1Device::PIO1Device()
	: inherited("PIO1")
{
	devaddr = 0x4d000000;
	devlen  = 4;

	poffevent.func = (DeviceCallback_t)&PIO1Device::PowerOffCallback;
	poffevent.SetName("PIO1 Power Off");
}

// デストラクタ
PIO1Device::~PIO1Device()
{
}

// リセット
void
PIO1Device::ResetHard()
{
	inherited::ResetHard();

	poffevent.Stop();
}

uint64
PIO1Device::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:
		// PIO1 PortA: LCD データ
		return gLCD->Read();

	 case 1:
		// PIO1 PortB: 無効
		return 0xff;

	 case 2:
		// PortC
		data = gLCD->Get();
		data |= xp_intreq ? 0 : 0x02;
		// 残りのビットは %1 だろうか?
		data |= 0x0d;
		putlog(1, "PC 読み込み -> $%02x", data);
		return data;

	 case 3:
		data = ReadCtrl();
		putlog(1, "コントロール読み込み -> $%02x", data);
		return data;

	 default:
		__unreachable();
	}
}

uint64
PIO1Device::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case 0:	// PIO1 PortA: LCD データ
		gLCD->Write(data);
		return 0;

	 case 1:	// PIO1 PortB: ローカル割り込み
		break;

	 case 2:	// PIO1 PortC: LCD/パワーオフ制御
		for (int i = 0; i < 8; i++) {
			SetPC(i, data & 1);
			data >>= 1;
		}
		return 0;

	 case 3:	// コントロール
		WriteCtrl(data);
		return 0;

	 default:
		__unreachable();
	}
	PANIC("$%08X <- $%02X 未サポート書き込み", gMPU->GetPaddr(), data);
	return 0;
}

uint64
PIO1Device::Peek(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case 0:	// PIO1 PortA LCD データ
		return gLCD->Peek();

	 case 1:	// PIO1 PortB 無効
		return 0xff;

	 case 2:	// PIO1 PortC
		data = gLCD->Get();
		data |= xp_intreq ? 0 : 0x02;
		// 残りのビットは %1 だろうか?
		data |= 0x0d;
		return data;

	 case 3:
		return ReadCtrl();

	 default:
		__unreachable();
	}
}

void
PIO1Device::SetPC(int pc, int val)
{
	switch (pc) {
	 case 0:
	 case 1:
	 case 2:
	 case 3:
		// PortC の下半分は入力ポートとして使われることを想定して
		// いるので、ここへの書き込みは起きないだろうし、無視する。
		return;

	 case 4:
		power = val;
		if (power) {
			// 電源オフ取り消し
			putlog(1, "電源 OFF 取消");
			poffevent.Stop();
		} else {
			// 1msec 後に電源オフ
			putlog(1, "電源 OFF 指示");
			poffevent.time = 1_msec;
			poffevent.Start();
		}
		return;
	 case 5:
		gLCD->SetRW(val);
		return;
	 case 6:
		gLCD->SetRS(val);
		return;
	 case 7:
		gLCD->SetE(val);
		return;

	 default:
		break;
	}
	PANIC("PC%d 未実装", pc);
}

// 電源 OFF イベントのコールバック
void
PIO1Device::PowerOffCallback(Event& ev)
{
	putlog(0, "電源 OFF");
	gScheduler->RequestPowerOff();
}
