//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// レンダラ
//

#pragma once

#include "device.h"
#include <condition_variable>
#include <mutex>

class Renderer : public Device
{
	using inherited = Device;
 public:
	// wxWidgets のビットマップは R8_G8_B8 の3バイト。
	static const int BPP = 3;

	static const uint REQ_TERMINATE		= 0x01;	// スレッド終了指示
	static const uint REQ_REND_POWEROFF	= 0x02;	// 電源オフ時の描画指示
	static const uint REQ_REND_NORMAL	= 0x04;	// 通常の描画指示
	static const uint REQ_INVALIDATE	= 0x08;	// 全画面更新指示

 public:
	Renderer();
	virtual ~Renderer() override;

	bool Init() override;
	bool PowerOff() override;

	// バッファとコールバック関数を指定
	// XXX 名前...
	void SetBuf(uint8 *, void (*)(void));

	// 画面合成処理 (CRTC とかから呼ばれる)
	void Render();

	// 全画面無効化の指示 (パレット変更で呼ばれる)
	void Invalidate();

	// 仮想画面幅(ピクセル)を取得
	int GetWidth() const { return width; }

	// 仮想画面高さ(ピクセル)を取得
	int GetHeight() const { return height; }

	// レンダリングスレッド (エントリポイントから呼ばれる)
	void ThreadRun();

 protected:
	// 画面合成。更新があれば true を返す。(派生クラス側で用意)
	virtual bool RenderMain() = 0;

	// ウィンドウとして表示される画面はオプションや設定などで縮尺を変える
	// ことが出来るが (こっちを実画面とする)、ここで扱う仮想画面とはそれに
	// 関わらず常にその機種の最大描画面積を指す。
	// LUNA でいう 1280 x 1024 のこと。

	// 仮想画面幅(ピクセル)
	int width {};

	// 仮想画面高さ(ピクセル)
	int height {};

	// バッファ (実体は wx 側で用意)
	uint8 *imagebuf {};

	// リフレッシュを指示するためのコールバック (wx 側で用意)
	// ここから直接 wxWindow::Refresh() を呼べないため。
	void (*refresh_callback)(void) {};

	// 次の画面更新で全画面を無効化する場合は true
	bool invalidate {};

	// リクエストフラグ
	uint32 request {};
	std::mutex mtx {};
	std::condition_variable cv {};

 private:
	pthread_t thread {};
	bool thread_created {};		// スレッドが作成されたら true

	// スレッドの終了を指示
	void Terminate();
};

//
// X68030 レンダラ
//
class X68030Renderer : public Renderer
{
	using inherited = Renderer;
 public:
	X68030Renderer();
	virtual ~X68030Renderer() override;

	bool RenderMain() override;
};

//
// LUNA レンダラ
//
class LunaRenderer : public Renderer
{
	using inherited = Renderer;
 public:
	LunaRenderer();
	virtual ~LunaRenderer() override;

	bool Init() override;

	bool RenderMain() override;

 private:
	const uint8 *pal {};
};

extern std::unique_ptr<Renderer> gRenderer;
