//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// LUNA シリーズの ROM エミュレーション共通部分
//
// IODevice
//    |
//    +-- ROMDevice (LoadROM()、ウェイト、マスク等を持つ)
//    |     |
//    |     +-- PROMDevice    (LUNA* の PROM)
//    |     +-- IPLROM1Device (X680x0 の IPLROM 後半)
//    |     +-- IPLROM2Device (X680x0 の IPLROM 前半)
//    |     +-- CGROMDevice   (X680x0 の CGROM)
//    |     |
//    |     |  +-------------------+
//    |     +--| LunaPROMEmuDevice | (LUNA PROM エミュレーションの共通部分)
//    |        +-------------------+
//    |           |
//    |           +-- Luna1PROMEmuDevice   (LUNA1 の PROM エミュレーション)
//    |           +-- Luna88kPROMEmuDevice (LUNA88K の PROM エミュレーション)
//    |
//    +-- PROM0Device   (LUNA* のブートページ切り替え用プロキシ)
//    +-- IPLROM0Device (X680x0 のブートページ切り替え用プロキシ)

#include "romemu_luna.h"
#include "bt454.h"
#include "lunafb.h"
#include "mainapp.h"
#include "mk48t02.h"
#include "ram.h"
#include "sio.h"
#include "spc.h"
#include "ufs.h"

// コンストラクタ
LunaPROMEmuDevice::LunaPROMEmuDevice()
	: inherited("PROM")
{
	// LUNA では ROM への書き込みは何も起きない
	write_op = 0;
}

// デストラクタ
LunaPROMEmuDevice::~LunaPROMEmuDevice()
{
}

bool
LunaPROMEmuDevice::Init()
{
	// 都度都度 dynamic_cast は面倒なので、ここで一度代入しとく
	nvram = dynamic_cast<MK48T02Device*>(gRTC.get());

	return true;
}

uint64
LunaPROMEmuDevice::Read32(uint32 addr)
{
	// ROM からのロングワードアクセスでだけ謎の I/O 空間が見える。
	// addr をマスクする前に評価すること。
	if ((gMPU->GetPaddr() & 0xff000000) == baseaddr) {
		switch (addr) {
		 case ROMIO_INIT:
			return ROM_Init();

		 case ROMIO_KEYIN:
			ROM_Keyin();
			return 0;

		 case ROMIO_CLOSE:
			ROM_Close();
			return 0;

		 case ROMIO_ENTRY:
			return ROM_Entry();

		 default:
			break;
		}
	}

	return inherited::Read32(addr);
}

// デバイスの初期化の共通部分。(継承クラスの ROM_Init() から呼ぶ)
void
LunaPROMEmuDevice::InitDevice()
{
	// 初期パレット
	// 偶数番は白(R/G/B=15/15/15)、奇数番は黒(0/0/0)。
	uint8 val = 0xf0;
	gBT454->Write(0, 0);	// select palette
	for (int i = 0; i < 16; i++) {
		gBT454->Write(1, val);	// R
		gBT454->Write(1, val);	// G
		gBT454->Write(1, val);	// B
		val ^= 0xf0;
	}

	// SCSI
	// BDID の書き込みによってホストデバイスがバスにアタッチされる構造なので。
	gSPC->Write(SPC::BDID, 7);

	// キーボード(chB)を初期化
	gSIO->Write(3, 0x01);		// CR1
	gSIO->Write(3, 0x10);		//    RX Int(all char)
	gSIO->Write(3, 0x03);		// CR3
	gSIO->Write(3, 0xc1);		//    RX Enable, 8bit
	gSIO->Write(3, 0x04);		// CR4
	gSIO->Write(3, 0x44);		//    Stop 1bit
	gSIO->Write(3, 0x05);		// CR5
	gSIO->Write(3, 0x68);		//    TX Enable, 8bit
}

// スクリーンを初期化
void
LunaPROMEmuDevice::InitScreen(const char *machine_name)
{
	text.Init(80, 32, TextScreen::Console);
	gLunafb->EmuConsoleOpen(&text);
	inputbuf.clear();
	inputpos = 0;
	is_shift = false;
	mousecnt = 0;

	text.Print(0, 0, "NONO %d.%d.%d Emulated ROM Monitor for %s",
		NONO_MAJOR_VER, NONO_MINOR_VER, NONO_PATCH_VER, machine_name);
	text.Locate(0, 2);
}

// ROM 処理のクローズ
void
LunaPROMEmuDevice::ROM_Close()
{
	// SIO をリセットしておく。
	// SIO のバッファにデータが残っていると、NetBSD のブートローダが
	// 起動してこないようだ。リセットして RXEN を落としておけば大丈夫ぽい。
	gSIO->ResetHard();

	// コンソールクローズ
	gLunafb->EmuConsoleClose();
}

// ROM プロンプトのキー入力。
// ゲストコードの割り込みハンドラから呼ばれて、キー入力で行を構成し、
// あるいは行が完成すればコマンドを実行する。
// 実行結果により画面を更新したりメンバ変数を更新する。
void
LunaPROMEmuDevice::ROM_Keyin()
{
	uint32 lunakey = gSIO->Read(2);

	// マウス入力の 2, 3バイト目は無視
	if (mousecnt > 0) {
		mousecnt--;
		return;
	}

	// SHIFT キーだけ状態を持つので先に処理
	if (lunakey == 0x0c || lunakey == 0x0d) {	// 左右SHIFT押下
		is_shift = true;
		return;
	}
	if (lunakey == 0x8c || lunakey == 0x8d) {	// 左右SHIFT開放
		is_shift = false;
		return;
	}
	// マウスデータの1バイト目なら、後続2バイトをスキップ
	if ((lunakey & 0xf8) == 0x80) {
		mousecnt = 2;
		return;
	}
	// キー開放は無視してよい
	if ((lunakey & 0x80)) {
		return;
	}

	// キーコードから文字を取得
	char asciicode;
	if (is_shift) {
		asciicode = lunakey2shifttable[lunakey];
	} else {
		asciicode = lunakey2asciitable[lunakey];
	}
	if (asciicode != 0) {
		GetChar(asciicode);
	}
}

// キー入力ハンドラ (キー入力割り込みによって呼ばれる)。
// 引数 asciicode は概ね ASCII コード (ただし 0x1c..0x1f が上下左右)。
void
LunaPROMEmuDevice::GetChar(char asciicode)
{
	// LUNA1 の場合、
	// 実機 PROM にはカーソル左右移動の機能はなく、カーソルは常に行末にいる。
	// そのためかどうか Delete も BackSpace と同じく1文字前を消す動作をして
	// いる。カーソル位置に文字はないので、Delete をそうするのはまあ分からん
	// でもない。
	// 一方、エミュレーション ROM では、カーソルの左右移動をサポートしてみた
	// ため Delete を「カーソル位置の文字を消す」にすることは出来るが、それは
	// それで実機と乖離が広がるので、どうしたもんか。

	switch (asciicode) {
	 case BS:
		if (inputbuf.empty() == false && inputpos != 0) {
			inputbuf.erase(--inputpos, 1);
		}
		break;

	 case LF:
		// 改行を起こす前にカーソル(ただの反転)を戻しておく必要がある
		text.SetAttr(TA::Normal);
		// 改行
		text.Putc(asciicode);
		prompt_y = text.GetY();
		// コマンド実行
		Command();
		// プロンプトを更新
		ClearPrompt();
		break;

	 case LEFT:
		if (inputpos > 0) {
			inputpos--;
		}
		break;
	 case RIGHT:
		if (inputpos < inputbuf.length()) {
			inputpos++;
		}
		break;

	 case UP:
	 case DOWN:
	 case 0:	// キー割り当てなし、これは来ないはずだが一応
		return;

	 default:	// 通常文字のはず
		inputbuf.insert(inputbuf.begin() + inputpos, asciicode);
		inputpos++;
		break;
	}
	ConsoleUpdate();
}

// 入力バッファを初期化する。
void
LunaPROMEmuDevice::ClearPrompt()
{
	prompt_y = text.GetY();
	// 入力行をクリア
	inputbuf.clear();
	inputpos = 0;
}

// 画面を更新
void
LunaPROMEmuDevice::ConsoleUpdate()
{
	int x;
	int y;

	// カーソルを描画する予定の行が最下行を越えていたらスクロール
	int len = prompt.length() + inputpos;
	y = prompt_y + (len / text.GetCol());
	if (y >= text.GetRow()) {
		text.ScrollUp();
		prompt_y--;
	}

	// プロンプト行から下を一旦全部クリア
	// (手抜きだが、見た目問題ないはず)
	text.Mode = TextScreen::Fixed;
	for (y = prompt_y; y < text.GetRow(); y++) {
		text.Locate(0, y);
		for (x = 0; x < text.GetCol(); x++) {
			text.Putc(' ');
		}
	}
	text.Locate(0, prompt_y);
	text.Mode = TextScreen::Console;

	// プロンプトと現在の入力行を出力
	text.Puts(prompt.c_str());
	text.Puts(inputbuf.c_str());

	// カーソルを描画
	y = prompt_y + (len / text.GetCol());
	x = len % text.GetCol();
	putmsg(2, "|%s|%s| pos=%d prompt_y=%d cur=(%d,%d)",
		prompt.c_str(), inputbuf.c_str(), inputpos, prompt_y, x, y);
	text.Locate(x, y);
	text.SetAttr(TA::On);

	// 画面更新
	gLunafb->EmuConsoleUpdate();
}

// NVRAM のチェックサムを返す
uint8
LunaPROMEmuDevice::CalcNVRAMCsum() const
{
	uint8 eor = 0;
	for (uint32 addr = 0x20; addr < 0x560; addr++) {
		eor ^= nvram->Peek(addr);
	}
	return eor;
}

// NVRAM のチェックサムを計算して書き込む
void
LunaPROMEmuDevice::WriteNVRAMCsum()
{
	uint8 eor = CalcNVRAMCsum();

	nvram->Write(0x001c, eor);
	nvram->Write(0x001d, eor);
	nvram->Write(0x001e, eor);
}

// NVRAM のマジックとチェックサムを照合する
bool
LunaPROMEmuDevice::VerifyNVRAM() const
{
	if (nvram->PeekString(0x0004) != "<nv>") {
		return false;
	}

	uint8 eor = CalcNVRAMCsum();
	for (uint32 addr = 0x1c; addr < 0x1f; addr++) {
		if (nvram->Peek(addr) != eor) {
			return false;
		}
	}
	return true;
}

// -X オプションで指定されたホストファイルをロードする。
// ロードできればエントリポイントを返す。
// exec_file が指定されている時に呼ぶこと。
uint32
LunaPROMEmuDevice::LoadHostFile()
{
	assert(gMainApp.exec_file);

	uint32 entry = gRAM->LoadFile(gMainApp.exec_file);
	if (entry) {
		text.Print("Host program loaded.  Entry point = $%08x\n",
			entry);
	} else {
		text.Print("** Couldn't load specified host program.\n");
	}
	return entry;
}

// bootinfo に従ってゲストから起動先をロードする。
// 成功すればエントリポイントを返す。
// 失敗ならをエラーメッセージを errmsg にセットして 0 を返す。
uint32
LunaPROMEmuDevice::LoadGuestFile(const bootinfo_t& bootinfo)
{
	const int& dkpart = bootinfo.dkpart;
	const int& dkunit = bootinfo.dkunit;
	const std::string& dkfile = bootinfo.dkfile;

	// ユニット (今は SCSI ID=6 決め打ち)
	int id = 6;
	SCSITarget *target = gSPC->GetTarget(id);
	if (target == NULL) {
		errmsg = "No bootable disks";
		putmsg(0, "%s", errmsg.c_str());
		return 0;
	}
	if (target->GetDevType() != SCSI::DevType::HD) {
		errmsg = "No bootable disks";
		putmsg(0, "%s", errmsg.c_str());
		return 0;
	}

	SCSIDisk *hd = dynamic_cast<SCSIDisk*>(target);
	putmsg(2, "%s SCSIHD[%d] found", __func__, id);

	// +0 セクタ目(512byte)に Sun disklabel
	scd_dk_label dk;
	if (hd->Peek(&dk, 0, sizeof(dk)) == false) {
		putmsg(0, "SCSIHD[%d] ReadSector failed", id);
		errmsg = string_format("dk%d Read disklabel failed.", dkunit);
		return 0;
	}

	// +$1fc の MAGIC をチェック
	uint32 magic = be16toh(dk.magic);
	if (magic != scd_dk_label::DKL_MAGIC) {
		putmsg(0, "SCSIHD[%d] Bad magic 0x%04x (!= 0x%x)",
			id, magic, scd_dk_label::DKL_MAGIC);
		errmsg = string_format("dk%d Bad disklabel magic.", dkunit);
		return 0;
	}

	// チェックサム照合は手抜きで省略

	// 指定パーティションの開始位置とサイズ
	if (loglevel >= 2) {
		// デバッグ用に全部表示
		for (int i = 0; i < 8; i++) {
			putmsgn("%s partition %c: { start=%8u size=%8u }", __func__,
				'a' + i,
				be32toh(dk.map[i].blkno),
				be32toh(dk.map[i].nblk));
		}
	}
	uint32 part_start = be32toh(dk.map[dkpart].blkno);
	uint32 part_size  = be32toh(dk.map[dkpart].nblk);
	putmsg(1, "%s partition=%c, start=%u size=%u", __func__,
		dkpart + 'a', part_start, part_size);

	// ファイルシステムをオープン (マウントっぽいイメージ)
	Filesys fsys(this);
	if (fsys.Mount(hd, part_start, part_size) == false) {
		errmsg = string_format("dk%d,%c Open ffs failed: %s",
			dkunit, dkpart + 'a', fsys.errstr.c_str());
		putmsg(0, "%s: %s", __func__, errmsg.c_str());
		return 0;
	}

	// ファイルを探してオープンする
	inodefile f(this);
	if (fsys.OpenFile(f, dkfile) == false) {
		errmsg = string_format("dk%d,%c,%s open failed: %s",
			dkunit, dkpart + 'a', dkfile.c_str(), fsys.errstr.c_str());
		putmsg(0, "%s: %s", __func__, errmsg.c_str());
		return 0;
	}

	// ファイル本体を読み込み
	fsys.ReadData(f);

	// ロード
	uint32 entry;
	std::string name = string_format("SCSIHD %d,%c:%s",
		id, dkpart + 'a', dkfile.c_str());
	entry = gRAM->LoadFile(name.c_str(), f.data.data(), f.data.size());
	if (entry == 0) {
		errmsg = string_format("dk%d,%c,%s not executable",
			dkunit, dkpart + 'a', dkfile.c_str());
	}

	return entry;
}

// キーコードから文字への変換
/*static*/ const uint8
LunaPROMEmuDevice::lunakey2asciitable[] = {
	 0,   0,   0,   0,   0,   0,   0,   0,
	 0,   0,   0,   0,   0,   0,   0,   0,
	 0,  BS,  LF,   0, ' ',   0,   0,   0,
	 0,   0,   0,   0,  UP,LEFT,RIGHT,DOWN,
	 0,   0,  '1', '2', '3', '4', '5', '6',
	'7', '8', '9', '0', '-', '^', '\\', 0,
	 0,   0,  'q', 'w', 'e', 'r', 't', 'y',
	'u', 'i', 'o', 'p', '@', '[',  0,   0,
	 0,   0,  'a', 's', 'd', 'f', 'g', 'h',
	'j', 'k', 'l', ';', ':', ']',  0,   0,
	 0,   0,  'z', 'x', 'c', 'v', 'b', 'n',
	'm', ',', '.', '/',  0,   0,   0,   0,
	 0, 0, 0, 0, 0, 0, 0, 0,	// XXX テンキー
	 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0,
};
/*static*/ const uint8
LunaPROMEmuDevice::lunakey2shifttable[] = {
	 0,   0,   0,   0,   0,   0,   0,   0,
	 0,   0,   0,   0,   0,   0,   0,   0,
	 0,  BS,  LF,   0, ' ',   0,   0,   0,
	 0,   0,   0,   0,  UP,LEFT,RIGHT,DOWN,
	 0,   0,  '!', '\"','#', '$', '%', '&',
	'\'','(', ')', ' ', '=', '~', '|',  0,		// XXX SHIFT+'0'は?
	 0,   0,  'Q', 'W', 'E', 'R', 'T', 'Y',
	'U', 'I', 'O', 'P', '`', '{',  0,   0,
	 0,   0,  'A', 'S', 'D', 'F', 'G', 'H',
	'J', 'K', 'L', '+', '*', '}',  0,   0,
	 0,   0,  'Z', 'X', 'C', 'V', 'B', 'N',
	'M', '<', '>', '?', '_',  0,   0,   0,
	 0, 0, 0, 0, 0, 0, 0, 0,	// XXX テンキー
	 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0,
};
