//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// LUNA-I の ROM エミュレーション
//
// IODevice
//    |
//    +-- ROMDevice (LoadROM()、ウェイト、マスク等を持つ)
//    |     |
//    |     +-- PROMDevice    (LUNA* の PROM)
//    |     +-- IPLROM1Device (X680x0 の IPLROM 後半)
//    |     +-- IPLROM2Device (X680x0 の IPLROM 前半)
//    |     +-- CGROMDevice   (X680x0 の CGROM)
//    |     +-- LunaPROMEmuDevice (LUNA PROM エミュレーションの共通部分)
//    |           |
//    |           |  +--------------------+
//    |           +--| Luna1PROMEmuDevice | (LUNA1 の PROM エミュレーション)
//    |           |  +--------------------+
//    |           |
//    |           +-- Luna88kPROMEmuDevice  (LUNA88K の PROM エミュレーション)
//    |
//    +-- PROM0Device   (LUNA* のブートページ切り替え用プロキシ)
//    +-- IPLROM0Device (X680x0 のブートページ切り替え用プロキシ)

#include "romemu_luna1.h"
#include "ethernet.h"
#include "lunafb.h"
#include "mainapp.h"
#include "memorystream.h"
#include "mk48t02.h"
#include "mpu680x0.h"
#include "pio.h"
#include "ram.h"

// コンストラクタ
Luna1PROMEmuDevice::Luna1PROMEmuDevice()
{
	// LUNA-I の ROM の特定アドレスを読み出す人がいるので、
	// サイズは実 ROM と同じにしておく必要がある。
	devlen = 128 * 1024;
	imagebuf.reset(new uint8 [devlen]);
	memset(&imagebuf[0], 0, devlen);
	mem = &imagebuf[0];

	mask = devlen - 1;
}

// デストラクタ
Luna1PROMEmuDevice::~Luna1PROMEmuDevice()
{
}

bool
Luna1PROMEmuDevice::Init()
{
	// 親クラス
	if (inherited::Init() == false) {
		return false;
	}

	MemoryStreamBE ms(imagebuf.get());
	ms.Write32(0x00002000);		// +00: リセットベクタ SP
	ms.Write32(0x41000400);		// +04: リセットベクタ PC
	ms.Write32(0);				// +08
	ms.Write32(0x41000500);		// +0c: メモリサイズ格納先

	// リセット時に実行する命令
	ms.SetPtr(&imagebuf[0x400]);
	ms.Write16(0x13c0);			//	move.b	d0,#0x41000000	; RAMに切り替え
	ms.Write32(0x41000000);
	ms.Write16(0x2039);			//	move.l	(ROMIO_INIT),d0
	ms.Write32(ROMIO_INIT);
	ms.Write16(0x660c);			//	bne		_go		; 起動先あればGO
								//_prompt:
	ms.Write32(0x4e722500);		//	stop	#0x2500	; キー入力を待つ
	ms.Write16(0x2039);			//	move.l	(ROMIO_ENTRY),d0
	ms.Write32(ROMIO_ENTRY);	//					; キー入力処理
	ms.Write16(0x67f4);			//	beq		_prompt	; 起動先なければループ
								//_go:
	ms.Write16(0x2040);			//	move.l	d0,a0	; ジャンプ先を a0 にセット
	ms.Write16(0x2039);			//	move.l	(ROMIO_CLOSE),d0
	ms.Write32(ROMIO_CLOSE);	//					; 後始末
	ms.Write16(0x4ed0);			//	jmp		(a0)	; ジャンプ

	// Lv6 (SIO) 割り込みハンドラ
	// キーボード割り込みが上がったら KEYIN をコールする
	ms.SetPtr(&imagebuf[0x600]);
	ms.Write16(0x4ab9);			//	tst.l	(ROMIO_KEYIN)
	ms.Write32(ROMIO_KEYIN);	//					; キー入力処理
	ms.Write16(0x4e73);			//	rte

	// MAC アドレス
	const macaddr_t macaddr = gEthernet->GetMacAddr();
	strcpy((char *)&imagebuf[0x1ffd8], "ENADDR");
	memcpy(&imagebuf[0x1ffe0], macaddr.to_string().c_str(), 12);

	// RAM にベクタを書き込む
	// XXX VBR を 41000000 に向けるのとどっちが楽かはある
	gRAM->Write32(0x78, 0x41000600);	// +78: Lv6 割り込み

	return true;
}

void
Luna1PROMEmuDevice::ResetHard()
{
	x_executed = false;
	x_entrypoint = 0;
}

// 起動時の処理を行う。
// 次段実行ファイルの読み込みに成功すればそのエントリポイントを返す。
// そうでなければ 0 を返すので、プロンプトに処理を移すこと。
uint32
Luna1PROMEmuDevice::ROM_Init()
{
	MemoryStreamBE ms;
	uint32 entry;

	// デバイスの初期化 (親クラス)
	InitDevice();

	// ROM 0x41000000.L リセットベクタ(SP,PC)もカーネルが参照している
	//                  (これは Init() が用意している)
	// ROM 0x4100000c.L メモリサイズ(バイト数)格納先アドレス
	// ROM 0x410000b8.L プレーンマスク格納先アドレス

	// メモリサイズはとりあえず 0x41000500 に格納
	ms.SetPtr(&imagebuf[0x500]);
	ms.Write32(gRAM->GetSize());

	// プレーン数(1..4)をプレーンマスク(0x01..0x0f)に
	int nplane = gLunafb->GetPlaneCount();
	int planemask = (1 << nplane) - 1;
	// 実際は RAM 上を指しているのだが、とりあえずすぐ隣に置いておく。
	ms.SetPtr(&imagebuf[0xb8]);
	ms.Write32(0x410000bc);
	ms.Write32(planemask);

	prompt_idx = 0;
	prompt = default_prompt;

	// スクリーンを初期化
	InitScreen("LUNA-I");

	entry = 0;

	// 表示環境が整ったところで NVRAM を照合
	if (VerifyNVRAM() == false) {
		InitNVRAM();
		text.Print("** NVRAM Initialized.\n");
		text.Print("\n");
		// NVRAM をクリアしたら DIPSW によらずプロンプトに降りる
		goto done;
	}

	// DIPSW#1-1 が UP なら自動起動、DOWN なら ROM モニタ起動なので、
	// これを真似する。
	if ((gPIO0->GetDIPSW1() & 1)) {
		if (gMainApp.exec_file) {
			// ホストのファイルを読み込んでエントリポイントアドレスを
			// この Read アクセスの読み出し値として返す。
			// エラーがあればエラーメッセージは呼び出し先側で表示してある。
			entry = LoadHostFile();
		} else {
			// NVRAM の設定に従ってブート先を探す。
			errmsg.clear();

			bootinfo_t bootinfo;
			GetNVRAM(bootinfo);
			entry = LoadGuestFile(bootinfo);
			if (entry == 0) {
				text.Print("** %s\n", errmsg.c_str());
			}
		}
	}

	// ここまでタイトルなどを Print してはいるが、実際の画面更新は
	// ConsoleUpdate() を呼ぶまで行われないことを利用(悪用?)して、
	// ここでエラーなくブートプログラムに制御を渡せる場合には何も表示せずに
	// ゲストに処理を移す。意図して或いは意図せずプロンプトに降りる場合だけ
	// ConsoleUpdate を呼ぶことでタイトルを含めて画面を表示する。
 done:
	if (entry == 0) {
		ClearPrompt();
		ConsoleUpdate();
	}

	return entry;
}

// ROM 処理。エントリポイントを返す。
uint32
Luna1PROMEmuDevice::ROM_Entry()
{
	if (x_executed && x_entrypoint != 0) {
		return x_entrypoint;
	} else {
		return 0;
	}
}

// ROM 処理のクローズ。
// 次段プログラムにジャンプするための処理も含む。
void
Luna1PROMEmuDevice::ROM_Close()
{
	inherited::ROM_Close();

	// NetBSD のブートローダがカーネルを読み込む時にはいくつかレジスタ
	// 渡しのパラメータがあるので、ここで用意する。本当はターゲットが
	// NetBSD カーネルの時に限定すべきのような気もするけど、とりあえず。
	// ROM がブートローダを読み込む時にはそのような取り決め事は一切ない
	// ので、ブートローダを読む時にも同じ値が入ってしまうが気にしない
	// ことにする。
	// 必要なパラメータは
	// %d6 に bootdev、NetBSD/luna68k の場合は
	// bootdev = $MACUPP00
	//            ||||++-- パーティション番号
	//            |||+---- SCSI ID (XXX ただしここでは 6 決め打ち)
	//            ||+----- unit (spc0 なら 0 のやつのはず)
	//            |+------ アダプタ (spc = 0、lance = 1)
	//            +------- magic ($a)
	//
	// %d7 に howto (0 でよい)
	// (と %a0 にジャンプ先を入れるがこれは Init() が書き出した ROM
	// アセンブラコード側で行っている)
	m68kcpu *cpu = gMPU680x0->GetCPU();
	RegD(6) = 0xa0060000;
	RegD(7) = 0;
}

// 入力コマンド処理 (入力行が揃ったところで呼ばれる)。
void
Luna1PROMEmuDevice::Command()
{
	std::vector<char> buf(inputbuf.length() + 1);
	std::vector<std::string> arg;

	// 1ワードごとに分解
	strcpy(&buf[0], inputbuf.c_str());
	char *p = &buf[0];
	for (char *ap; (ap = strsep(&p, " ")) != NULL; ) {
		if (*ap != '\0') {
			arg.push_back(std::string(ap));
		}
	}
	if (loglevel >= 1) {
		for (int i = 0; i < arg.size(); i++) {
			putmsgn("arg[%d]=|%s|", i, arg[i].c_str());
		}
	}

	// "k" コマンド中ならここで先に処理する
	if (prompt_idx != 0) {
		CommandK(arg);
		return;
	}

	// 空行
	if (arg.size() == 0) {
		return;
	}

	// 1ワード目がコマンド
	if (arg[0] == "h") {
		// ヘルプメッセージ
		text.Print("Commands\n");
		text.Print(
"  g [<filename>]     : load program\n"
"  gh                 : load host program (specified by -A option)\n"
"  h                  : show this message\n"
"  k                  : show/set boot constants\n"
"  x                  : execute loaded program\n"
		);
		return;
	}
	if (arg[0] == "g") {
		CommandG(arg);
		return;
	}
	if (arg[0] == "gh") {
		CommandGH();
		return;
	}
	if (arg[0] == "k") {
		CommandK(arg);
		return;
	}
	if (arg[0] == "x") {
		CommandX();
		return;
	}

	text.Print("** Unknown command: %s\n", arg[0].c_str());
}

// "g" コマンド
// g なら NVRAM の情報に従ってプログラムをロード。
// g <fname> なら NVRAM の fname だけ指定の名前を使ってロード。NVRAM には
// 書き込まない。
void
Luna1PROMEmuDevice::CommandG(const std::vector<std::string>& arg)
{
	std::string ctlr = nvram->PeekString(0x0030);
	if (ctlr == "dk") {			// harddisk
		std::string fname {};

		switch (arg.size()) {
		 case 1:
			break;
		 case 2:
			fname = arg[1];
			break;
		 default:
			text.Print("** Invalid argument.\n");
			return;
		}
		errmsg.clear();

		bootinfo_t bootinfo;
		GetNVRAM(bootinfo);
		// 引数で指定されたらファイル名だけ上書き
		if (!fname.empty()) {
			bootinfo.dkfile = fname;
		}
		x_entrypoint = LoadGuestFile(bootinfo);
		if (x_entrypoint == 0) {
			text.Print("** %s\n", errmsg.c_str());
		}

	} else if (ctlr == "et" ||	// netboot
	           ctlr == "sd" ||	// cassette tape
	           ctlr == "fl"   )	// floppy
	{
		text.Print("** %s boot not supported.\n", ctlr.c_str());
		return;
	} else {
		// 起きないはず
		text.Print("** Invalid boot controller name.\n");
		return;
	}

	if (x_entrypoint) {
		text.Print("Loaded.  Entry point = $%08x\n", x_entrypoint);
	}
}

// "gh" コマンド (勝手に拡張)
// -X オプションで指定されたホストプログラムをロードする。
void
Luna1PROMEmuDevice::CommandGH()
{
	// 指定されてたら読み込む
	if (gMainApp.exec_file == NULL) {
		text.Print("** Cannot loaded; host program is not specified.\n");
		return;
	}
	x_entrypoint = LoadHostFile();
}


// k コマンドのパラメータ表
struct kinfo {
	uint32 offset;				// NVRAM 上のオフセット
	const char * const name;	// 表示名
	int next;					// 次インデックス
};
static const struct kinfo kinfo[] = {
	{ 0 },	// 0 は通常プロンプトを指すのでここでは使わない
	{ 0x0030,	"controller",  0 },	// [1] BOOT (next は動的判定する)
	{ 0x0090,	"drive unit",  3 },	// [2] DKUNIT
	{ 0x00b0,	"partition ",  4 },	// [3] DKPART
	{ 0x00d0,	"filename  ",  0 },	// [4] FLNAME
	{ 0x0050,	"hostname  ",  6 },	// [5] HOST
	{ 0x0070,	"servername",  7 },	// [6] SERVER
	{ 0x0190,	"filename  ",  8 },	// [7] ETBOOT
	{ 0x01b0,	"",			   0 },	// [8] ETFILE (ETBOOT と1行で処理する)
};

// "k" コマンド
void
Luna1PROMEmuDevice::CommandK(const std::vector<std::string>& arg)
{
	uint32 offset;
	const char *name;
	int next;

	// 入力を処理
	if (prompt_idx == 0) {
		// ここは1行目表示前なので前行の入力はない。
		next = 1;
	} else {
		std::string val;
		bool rv;

		// 入力があれば更新
		if (arg.size() != 0) {
			rv = CommandKInput(arg[0]);
		} else {
			// [7] の時に入力があれば CommandKInput() の中で更新されるが、
			// 入力がなければそこを通らないので、こっちでやらないといけない。
			// うーん。
			if (prompt_idx == 7) {
				prompt_idx = kinfo[prompt_idx].next;
			}
			rv = true;
		}

		// これ以降は更新後の prompt_idx を使う

		// 入力エラーがあった時はこの行を再入力するため
		next = prompt_idx;

		if (rv == true) {
			// 入力があっても空でも次インデックスへ。
			 if (prompt_idx == 1) {
				// BOOT なら現在値によって次のインデックスが変わる
				// "dk" なら [2]、"et" なら [5]。他は未対応
				offset = kinfo[prompt_idx].offset;
				val = nvram->PeekString(offset);
				if (val == "dk") {
					next = 2;
				} else if (val == "et") {
					next = 5;
				} else {
					// 起きることはないはず
					text.Print("** Inconsistent boot parameter: %s\n",
						val.c_str());
				}
			} else {
				// それ以外はテーブルにかかれている通り
				next = kinfo[prompt_idx].next;
			}
		}
	}

	// この行のプロンプト文字列を作成
	prompt_idx = next;
	if (prompt_idx == 0) {
		// 最終行なら元に戻す
		prompt = default_prompt;
	} else {
		// 次のプロンプトに変更
		name = kinfo[prompt_idx].name;
		offset = kinfo[prompt_idx].offset;
		std::string val = nvram->PeekString(offset);

		// [7] ETBOOT だけ1行で ETBOOT:ETFILE 形式
		if (prompt_idx == 7) {
			next = kinfo[prompt_idx].next;	// ETFILE
			offset = kinfo[next].offset;
			val += ":";
			val += nvram->PeekString(offset);
		}

		prompt = string_format("%s: %s  ?", name, val.c_str());
	}
}

// k コマンドの入力値を処理する部分。
// 成功すれば true を返す。
// 失敗すれば false を返す。
// 実機 ROM は入力値にエラーがあると黙って再入力を促す。
// エラーメッセージくらい表示してもいいかもだがとりあえず真似しておく。
bool
Luna1PROMEmuDevice::CommandKInput(const std::string& val)
{
	int offset;
	bool rv = true;

	offset = kinfo[prompt_idx].offset;

	switch (prompt_idx) {
	 case 1:	// BOOT
		// 正常な入力値は "dk", "et", "sd", "fl" のみ。
		if (val != "dk" && val != "et" && val != "sd" && val != "fl") {
			return false;
		}
		rv = nvram->WriteString(offset, val);
		if (rv == false) {
			return false;
		}
		break;

	 case 7:	// ETBOOT
	 {
		// こいつだけ ETBOOT:ETFILE 形式で2つのパラメータを同時に処理する

		// ':' で前後に分ける
		int sep = val.find(':');
		if (sep == std::string::npos) {
			return false;
		}
		std::string etboot = val.substr(0, sep);
		std::string etfile = val.substr(sep + 1, val.size() - sep + 1);
		if (etboot.empty() || etfile.empty()) {
			return false;
		}

		// ETBOOT
		rv = nvram->WriteString(offset, etboot);
		if (rv == false) {
			return false;
		}

		// 次へ
		prompt_idx = kinfo[prompt_idx].next;

		// ETFILE
		offset = kinfo[prompt_idx].offset;
		rv = nvram->WriteString(offset, etfile);
		if (rv == false) {
			return false;
		}
		break;
	 }

	 default:
		rv = nvram->WriteString(offset, val);
		if (rv == false) {
			return false;
		}
		break;
	}

	WriteNVRAMCsum();
	return rv;
}

// "x" コマンド
// 読み込まれていたらフラグを立てるだけ。
void
Luna1PROMEmuDevice::CommandX()
{
	if (x_entrypoint == 0) {
		text.Print("** Program is not loaded.\n");
	} else {
		x_executed = true;
	}
}

// NVRAM を初期化する
bool
Luna1PROMEmuDevice::InitNVRAM()
{
	nvram->ClearAll();

	// 初期値どうしたもんか
	nvram->WriteString(0x0004, "<nv>");
	nvram->WriteString(0x0020, "BOOT");
	nvram->WriteString(0x0030, "dk");
	nvram->WriteString(0x0040, "HOST");
	nvram->WriteString(0x0050, "nono");	// 偽札的なアレ
	nvram->WriteString(0x0060, "SERVER");
	nvram->WriteString(0x0070, "servername");
	nvram->WriteString(0x0080, "DKUNIT");
	nvram->WriteString(0x0090, "0");
	nvram->WriteString(0x00a0, "DKPART");
	nvram->WriteString(0x00b0, "c");
	nvram->WriteString(0x00c0, "DKFILE");
	nvram->WriteString(0x00d0, "vmunix");
	nvram->WriteString(0x00e0, "FLUNIT");
	nvram->WriteString(0x00f0, "2");
	nvram->WriteString(0x0100, "FLFILE");
	nvram->WriteString(0x0110, "vmunix");
	nvram->WriteString(0x0120, "STUNIT");
	nvram->WriteString(0x0130, "0");
	nvram->WriteString(0x0140, "STFLNO");
	nvram->WriteString(0x0150, "1");
	nvram->WriteString(0x0160, "STFILE");
	nvram->WriteString(0x0170, "vmunix");
	nvram->WriteString(0x0180, "ETBOOT");
	nvram->WriteString(0x0190, "server");
	nvram->WriteString(0x01a0, "ETFILE");
	nvram->WriteString(0x01b0, "/vmunix");
	nvram->WriteString(0x0560, "ENADDR");
	nvram->WriteString(0x0570, "");	// どうするか

	// チェックサム再計算
	WriteNVRAMCsum();

	return true;
}

// NVRAM のブート設定を bootinfo に取得する。
void
Luna1PROMEmuDevice::GetNVRAM(bootinfo_t& bootinfo)
{
	// 起動先情報を NVRAM (mk48t02) に問い合わせる
	// DKUNIT が装置番号(今はまだ SCSI ID=6 固定)。
	// DKPART が 'a'..'h' でパーティション。
	// DKFILE が読み込む次段プログラム (NetBSD では通常 /boot のこと)。
	bootinfo.dkunit = nvram->Peek(0x0090) - '0';
	bootinfo.dkpart = nvram->Peek(0x00b0) - 'a';
	bootinfo.dkfile = nvram->PeekString(0x00d0);
	putmsg(1, "%s DKUNIT=%d", __func__, bootinfo.dkunit);
	putmsg(1, "%s DKPART=%c", __func__, bootinfo.dkpart + 'a');
	putmsg(1, "%s DKFILE=|%s|", __func__, bootinfo.dkfile.c_str());
}
