//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "sram.h"
#include "mainapp.h"

//
// SRAM
//
// SRAM はビッグエンディアンのままメモリに置かれるので、
// 必要ならアクセス時に都度変換すること。
//

std::unique_ptr<SRAMDevice> gSRAM;

SRAMDevice::SRAMDevice()
	: inherited("SRAM")
{
	devaddr = baseaddr;
	devlen  = 0x4000;
}

SRAMDevice::~SRAMDevice()
{
}

bool
SRAMDevice::Init()
{
	filename = gMainApp.GetVMDir() + "SRAM030.DAT";
	file.SetFilename(filename);
	file.SetDispname(string_format("SRAM \"%s\"", filename.c_str()));
	mem = file.OpenCreate(16 * 1024);
	if (mem == NULL) {
		return false;
	}

	return true;
}

void
SRAMDevice::ResetHard()
{
	// XXX 未調査
	writeable = false;
}

// アドレスデコーダ
inline uint64
SRAMDevice::Decoder(uint32 addr) const
{
	uint32 offset = addr - baseaddr;
	assertmsg(offset < file.GetMemSize(),
		"offset=%x memlen=%x", offset, (int)file.GetMemSize());
	return offset;
}

uint64
SRAMDevice::Read8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	uint32 data;

	// ホストファイル起動モードなら起動アドレスとして Pluto-X を見せる。
	if (gMainApp.exec_file) {
		data = PeekHostBoot(offset);
		if ((int32)data >= 0) {
			if ((offset & 1) == 0) {
				data = data >> 8;
			} else {
				data = data & 0xff;
			}
			return data;
		}
	}

	data = mem[offset];
	return data;
}

uint64
SRAMDevice::Read16(uint32 addr)
{
	uint32 offset = Decoder(addr);
	uint32 data;

	// ホストファイル起動モードなら起動アドレスとして Pluto-X を見せる。
	if (gMainApp.exec_file) {
		data = PeekHostBoot(offset);
		if ((int32)data >= 0) {
			return data;
		}
	}

	data = be16toh(*(uint16 *)&mem[offset]);
	return data;
}

uint64
SRAMDevice::Write8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);

	if (writeable) {
		if (offset < 0x100) {
			putlog(1, "更新 $%06x <- $%02x", addr, data);
		}
		mem[offset] = data;
	} else {
		putlog(1, "書き込み禁止 $%06x <- $%02x", addr, data);
	}
	return 0;
}

uint64
SRAMDevice::Write16(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);

	if (writeable) {
		if (offset < 0x100) {
			putlog(1, "更新 $%06x <- $%04x", addr, data);
		}
		*(uint16 *)&mem[offset] = htobe16(data);
	} else {
		putlog(1, "書き込み禁止 $%06x <- $%04x", addr, data);
	}
	return 0;
}

uint64
SRAMDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);

	// ホストファイル起動モードなら起動アドレスとして Pluto-X を見せる。
	if (gMainApp.exec_file) {
		uint32 data = PeekHostBoot(offset);
		if ((int32)data >= 0) {
			if ((offset & 1) == 0) {
				data = data >> 8;
			} else {
				data = data & 0xff;
			}
			return data;
		}
	}

	return mem[offset];
}

void
SRAMDevice::write_enable(bool value)
{
	if (writeable == false && value == true) {
		putlog(1, "書き込み許可");
	} else if (writeable == true && value == false) {
		putlog(1, "書き込み禁止");
	}

	writeable = value;
}

// ホストファイル起動用の読み出し値を返す。
// offset はワード境界であること。戻り値はホストバイトオーダのワード値。
// 本来の SRAM の値を読み出してよいところについては (uint32)-1 を返す。
//
// mem[] は SRAM.DAT を直接 mmap(2) しており書き換えておくことは出来ないため、
// ホストファイル起動モードなら読み出し時に動的に値を返す。
uint32
SRAMDevice::PeekHostBoot(uint32 offset) const
{
	switch (offset) {
	 case 0x0c:			// ROM 起動アドレス(H)
		return (plutoaddr >> 16);
	 case 0x0e:			// ROM 起動アドレス(L)
		return (plutoaddr & 0xffff);
	 case 0x18:			// 起動デバイスを ROM にする
		return 0xa000;
	 default:
		break;
	}
	return (uint32)-1;
}
