//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "vm.h"
#include "bus.h"
#include "mainapp.h"
#include "mpu.h"
#include "mythread.h"
#include "ram.h"
#include "renderer.h"
#include "rtc.h"
#include "scheduler.h"

std::unique_ptr<VM> gVM;

// コンストラクタ
VM::VM()
{
	// どこでやるのがいいかわからんけど、このコンストラクタはメインスレッドで
	// 作られるので、ここでメインスレッドの名前を設定。
	PTHREAD_SETNAME("Main");

	// バスエラーデバイスだけここで作成。
	gBusErr.reset(new BusErrDevice());

	// まず全域をバスエラーで初期化。
	// この後継承先で必要に応じて上書きする。
	for (int i = 0; i < countof(::devtable); i++) {
		::devtable[i] = gBusErr.get();
	}
}

// デストラクタ
VM::~VM()
{
	Dispose();
}

// 終了
void
VM::Dispose()
{
	if ((bool)gScheduler) {
		// スケジューラを停止。
		// スケジューラ(VMスレッド)からのアクセスさえなくなれば
		// 後の有象無象はグローバルデストラクタにお任せでもいいか。
		// だめだったらもっと頑張る。
		gScheduler->Terminate();
		gScheduler.reset();
	}
}

// コンストラクタ後の初期化
bool
VM::Create()
{
	// d->Create() は gDevices に要素を追加する場合があるため、ここで
	// ranged for は使えない。また追加されたデバイスの Create() も(念のため)
	// 実行したいため、ループ終了条件で毎回 size() を取り出して比較する。
	for (int i = 0; i < gDevices.size(); i++) {
		Device *d = gDevices[i];
		if (!d->Create()) {
			return false;
		}
	}

	return true;
}

// 初期化
bool
VM::Init()
{
	for (const auto& d : gDevices) {
		if (!d->Init()) {
			return false;
		}
	}

	return true;
}

// 設定の適用
bool
VM::Apply()
{
	for (const auto& d : gDevices) {
		if (!d->Apply()) {
			return false;
		}
	}

	return true;
}

// ブートページを切り替える (共通処理)
// 派生クラスから呼ぶこと。
void
VM::SwitchBootPage(Device *parent, bool isrom)
{
	// ログ表示
	if (isrom) {
		parent->putlogf(1, lstr("SwitchBootPage ROM (Boot)"));
	} else {
		parent->putlogf(1, lstr("SwitchBootPage RAM (Normal)"));
	}

	// スケジューラに指示
	gScheduler->RequestBootPageMode(isrom);
}
