//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "vm_x68k.h"
#include "bus.h"
#include "config.h"
#include "dmac.h"
#include "fdc.h"
#include "human68k.h"
#include "interrupt.h"
#include "mainapp.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "pedec.h"
#include "pluto.h"
#include "pio.h"
#include "renderer.h"
#include "rp5c15.h"
#include "romimg_x68k.h"
#include "scheduler.h"
#include "spc.h"
#include "x68kkbd.h"
#include "xiospace.h"

//
// X68030 VM
//

// コンストラクタ
VM_X68030::VM_X68030()
{
	// 機種固有パラメータ
	// MPU クロックの初期値は機種ごとに異なる。単位は MHz
	gConfig->SetDefault("mpu-clock", 25);
	// RAM 初期値は MAX にしておく
	gConfig->SetDefault("ram-size", 12);
	// SPC の入力クロック (5MHz = 200ns)
	gConfig->Add("!spc-tCLF", 200);

	// この機種に不要なパラメータはここで削除する
	// (機種が決まるより前に全機種の設定の和集合を用意してあるため)
	gConfig->Delete("luna-dipsw1");
	gConfig->Delete("luna-dipsw2");
	gConfig->Delete("prom-image");

	// デバイス (順序に注意)
	// BusErrDevice は親コンストラクタで作成済み。

	gMPU.reset(new MPU680x0Device());				// required by Scheduler
	gScheduler.reset(new Scheduler());				// required by *
	gInterrupt.reset(new X68030Interrupt());
	gRAM.reset(new RAMDevice());					// required by XIOSpace
	gXIOSpace.reset(new XIOSpaceDevice());
	gIPLROM0.reset(new IPLROM0Device());

	gKeyboard.reset(new X68030Keyboard());
	gRenderer.reset(new X68030Renderer());

	if (gMainApp.human_mode) {
		human68k.reset(new Human68k());
	}

	// 割り当て
	// 今は拡張メモリがないので、今は上位8ビット全領域がミラーになっている。
	for (int i = 0; i < countof(::devtable); i++) {
		::devtable[i] = gXIOSpace.get();
	}
}

// デストラクタ
VM_X68030::~VM_X68030()
{
}

// X68030 の電源ボタンを押す。
void
VM_X68030::PowerButton()
{
	// 電源	ボタン
	// ON	OFF->ON		電源オフ割り込みをネゲート?
	// ON	ON ->OFF	電源オフ割り込みをアサート?
	// OFF	OFF->ON		電源オン
	// OFF	ON ->OFF	(通常はあり得ないはず)

	// ボタンの状態を更新して..
	power_button = !power_button;

	// スケジューラに指示
	gScheduler->RequestPowerOn();
}

// ブートページを切り替える。
void
VM_X68030::SwitchBootPage(Device *parent, bool isrom)
{
	if (gMainApp.human_mode) {
		// Human モードではブートページを RAM に見せる。
		// 今のところ。
		return;
	}

	// 共通処理
	inherited::SwitchBootPage(parent, isrom);

	// X68030 では XIOSpace が担当している
	gXIOSpace->SwitchBootPage(isrom);
}

// リセットについて
//
//           PWRRESET
// 電源オン ----------+--> VIPS
// (7505)             +--> VideoCtlr
//                    +--> OSCIAN2
//                    |
//                    |   +------+
//                    +-->|      |
// 本体      SWRESET      | YUKI | IPLRESET
// リセット ------------->|      |-----------> FPU
// スイッチ               +------+
//                         |
//                         ▽
//                   RESET |
//                         |<-> MPU (68030)
//                         +--> PPI
//                         +--> PEDEC
//                         +--> SPC
//                         +--> ADPCM
//                         +--> FDC
//                         +--> MFP
//                         |
//                         |      EXRESET
//                         +---|>-----------> 外部バス
//                         ▽
//                SYSRESET |
//                         |<-> SAKI  ---+
//                         |             | BEC0,BEC1
//                         +--> DMAC  <--+
//                         +--> CYNTHIA
// リセットされない?
// o CACHY
// o YM2151 (FM音源)
// o SCC

// MPU の RESET 命令によるリセット
void
VM_X68030::ResetByMPU()
{
	gDMAC->ResetHard();
	gFDC->ResetHard();
	gMFP->ResetHard();
	gPEDEC->ResetHard();
	gPluto->ResetHard();
	gPPI->ResetHard();
	gSPC->ResetHard();
}
