//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxhistmonitor.h"

//
// 履歴モニタ (共通部分)
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXHistoryMonitor, inherited)
	EVT_SIZE(WXHistoryMonitor::OnSize)
	EVT_SCROLL(WXHistoryMonitor::OnScroll)
wxEND_EVENT_TABLE()

// コンストラクタ
WXHistoryMonitor::WXHistoryMonitor(wxWindow *parent, const wxString& name,
	Monitor& monitor_)
	: inherited(parent, wxID_ANY, name)
{
	// →
	// +--+-----------------------+--+--+
	// |  |↓+------------------+ |  |  |
	// |  |  | 上余白(padding1) | |  |  |
	// |左|  +------------------+ |右|  |
	// |余|  | TextScreen       | |余| <- VScroll
	// |白|  +------------------+ |白|  |
	// |  |  | 下余白(padding2) | |  |  |
	// |  |  +------------------+ |  |  |
	// +--+-----------------------+--+--+

	wxBoxSizer *topsizer = new wxBoxSizer(wxHORIZONTAL);
	// 左余白
	topsizer->Add(new WXPaddingPanel(this), 0, wxEXPAND);
	// 中列の縦 sizer
	auto *vbox = new wxBoxSizer(wxVERTICAL);
	topsizer->Add(vbox, 0, wxEXPAND);

	// パディングパネル(上下)
	// ウィンドウを1行ずつ伸縮させるためには、クライアント領域がフォント高さの
	// 整数倍になってないといけない。TextScreen はそれ自身が行の整数倍の大きさ
	// だが、上下の余白はその時点のフォントサイズによって変更する必要がある。
	// 先にサイズ不定のパネルだけ置いといて、フォントサイズ確定後に大きさを
	// 確定させる。DoSize() 参照。
	padding1 = new WXPaddingPanel(this);
	padding2 = new WXPaddingPanel(this);
	screen = new WXTextScreen(this, monitor_);
	vbox->Add(padding1, 0, wxEXPAND);
	vbox->Add(screen, 1, wxEXPAND);
	vbox->Add(padding2, 0, wxEXPAND);

	// 右余白
	topsizer->Add(new WXPaddingPanel(this), 0, wxEXPAND);

	// スクロールバー
	vscroll = new wxScrollBar(this, wxID_ANY, wxDefaultPosition,
		wxDefaultSize, wxSB_VERTICAL);
	topsizer->Add(vscroll, 0, wxEXPAND);

	SetSizer(topsizer);

	// ウィンドウサイズを確定させる
	DoSize();

	// スクロールのために (パネル領域での) MouseWheel イベントもここで
	// 受け取りたい。スクロールバー上のマウスホイール操作は最初から Scroll
	// イベントとして処理されているので問題ないが、パネル領域でのマウス
	// ホイール操作はここ(wxFrame)ではなくパネルに対して飛んでくる。
	// ここで一緒に処理したほうが楽なので、こちらに回す。
	screen->Connect(wxEVT_MOUSEWHEEL,
		wxMouseEventHandler(WXHistoryMonitor::OnMouseWheel), NULL, this);
}

// デストラクタ
WXHistoryMonitor::~WXHistoryMonitor()
{
}

// フォントサイズが確定したら、それによってウィンドウサイズを再計算する。
void
WXHistoryMonitor::DoSize()
{
	CalcVSize(0);
}

// サイズ変更イベント
void
WXHistoryMonitor::OnSize(wxSizeEvent& event)
{
	inherited::OnSize(event);

	// スクロールバーを再設定
	int pos = (int)screen->GetUserData();
	int thumbsize = screen->GetRow() - 1; // 常にヘッダ1行表示している
	int range = 256;	// XXX 拾ってくるべき
	int pagesize = thumbsize - 1;
	if (pos > range - thumbsize) {
		pos = range - thumbsize;
	}
	vscroll->SetScrollbar(pos, thumbsize, range, pagesize);
}

// マウスホイールイベント (WXTextScreen 上で起きたやつをこっちに回してある)
void
WXHistoryMonitor::OnMouseWheel(wxMouseEvent& event)
{
	// GetWheelRotation() は上(奥)向きに回すと +120、下(手前)に回すと -120。
	// どこの決まりか知らんけど、スクロールバーのほうはホイール1段階で3行
	// スクロールするのでそれに合わせる。
	int pos = (int)screen->GetUserData();
	pos = pos - event.GetWheelRotation() / 40;

	int maxpos = vscroll->GetRange() - vscroll->GetThumbSize();
	if (pos < 0)
		pos = 0;
	if (pos > maxpos)
		pos = maxpos;

	DoScroll(pos);
}

// スクロールイベント (全部入り)
void
WXHistoryMonitor::OnScroll(wxScrollEvent& event)
{
	DoScroll(event.GetPosition());
}

// スクロール処理 (OnScroll() と OnMouseWheel() から呼ばれる)
void
WXHistoryMonitor::DoScroll(int pos)
{
	// BranchHistory モニタは userdata が表示開始位置になっている
	screen->SetUserData(exflag | pos);
	vscroll->SetThumbPosition(pos);
}

//
// ブランチ履歴モニタ
//

// コンストラクタ
WXBranchHistoryMonitor::WXBranchHistoryMonitor(wxWindow *parent,
	Monitor& monitor)
	: inherited(parent, _("Branch History"), monitor)
{
	exflag = 0;
	screen->SetUserData(exflag);
}

// デストラクタ
WXBranchHistoryMonitor::~WXBranchHistoryMonitor()
{
}

//
// 例外履歴モニタ
//

// コンストラクタ
WXExceptionHistoryMonitor::WXExceptionHistoryMonitor(wxWindow *parent,
	Monitor& monitor)
	: inherited(parent, _("Exception History"), monitor)
{
	exflag = 1ULL << 63;
	screen->SetUserData(exflag);
}

// デストラクタ
WXExceptionHistoryMonitor::~WXExceptionHistoryMonitor()
{
}
