//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

// ステータスパネル
//
// 描画領域に対してフォントの開始位置は外枠から +4、内枠からなら +2。
// 横も同様。
//
// +0 ------ 外枠
// +1        アキ
// +2  +---- 内枠
// +3  |     アキ
// +4  | ここからフォント

#include "wxstatuspanel.h"
#include "wxcolor.h"
#include "wxmainframe.h"
#include "wxmainview.h"
#include "wxuimessage.h"
#include "wxversion.h"
#include "netdriver.h"
#include "scheduler.h"
#include "spc.h"

WXStatusPanel *gStatusPanel;

wxBEGIN_EVENT_TABLE(WXStatusPanel, inherited)
	EVT_CLOSE(WXStatusPanel::OnClose)
	EVT_SHOW(WXStatusPanel::OnShow)
	EVT_SIZE(WXStatusPanel::OnSize)
	EVT_TIMER(wxID_ANY, WXStatusPanel::OnTimer)
	EVT_PAINT(WXStatusPanel::OnPaint)
	EVT_LEFT_DCLICK(WXStatusPanel::OnLeftDClick)
	EVT_CONTEXT_MENU(WXStatusPanel::OnContextMenu)
wxEND_EVENT_TABLE()

// ステータスパネルの更新頻度 [Hz]
static constexpr uint FPS = 10;

// style フラグ
// デフォルトは wxTAB_TRAVERSAL だがこれを外すので分かりづらい字面になる
static const long STATUS_PANEL_STYLE = 0;

// コンストラクタ
WXStatusPanel::WXStatusPanel(wxWindow *parent)
	: inherited(parent, STATUS_PANEL_STYLE)
{
	timer.SetOwner(this);

	// インジケータ用のパラメータを用意。
	InitIndicators();

	DoSize();

	// キー入力イベントをメインビューに転送。
	// このパネルへのキー入力は不要なのと、メインウィンドウ内のコントロール
	// フォーカスがこのステータスパネルとメインビューのどっちにあるかは
	// 分かりづらくて、こっちがフォーカス持ってしまうと (というか起動後は
	// こっちにある) VM にキー入力が出来ずに焦ることになるので、ここへの
	// キー入力イベントは全部メインビューに飛ばす。
	Connect(wxEVT_KEY_UP, wxKeyEventHandler(WXMainView::OnKeyUp), NULL,
		gMainView);
	Connect(wxEVT_KEY_DOWN, wxKeyEventHandler(WXMainView::OnKeyDown), NULL,
		gMainView);

	// VM からのメディア変更通知を受け取る
	WXUIMessage::Connect(UIMessage::SCSI_MEDIA_CHANGE, this,
		wxCommandEventHandler(WXStatusPanel::OnSCSIMediaChanged));
}

// デストラクタ
WXStatusPanel::~WXStatusPanel()
{
}

// クローズイベント
void
WXStatusPanel::OnClose(wxCloseEvent& event)
{
	// クローズする前に更新を止める。
	timer.Stop();
}

// フォントサイズ変更 (外部インタフェース)
void
WXStatusPanel::SetFontSize(FontId new_fontid)
{
	// 同じなら何もしない
	if (fontid == new_fontid) {
		return;
	}

	// フォントを変更
	UpdateFont(new_fontid);

	// サイズ変更処理
	DoSize();
}

// 表示/非表示変更イベント
// new した時は (デフォルトで表示状態なのでか状態に変更がないという扱いの
// ようで) イベントは飛んでこない。同様に new 直後に Show() を明示発行しても
// 状態に変更がないためイベントは飛んでこない。
// new 直後に Hide() すればイベントは飛んでくる。という動作をするようだ。
// そのため WXMainFrame 側で生成後に Show/Hide を起こしている。
void
WXStatusPanel::OnShow(wxShowEvent& event)
{
	if (event.IsShown()) {
		// 非表示→表示

		// 表示開始前に一度情報を取得
		UpdateStat();

		timer.Start(1000 / FPS);
	} else {
		// 表示→非表示

		timer.Stop();
	}
}

// サイズ変更処理
// サイズを変更したい時に呼ぶ (サイズが変更されたからではない)。
// ここではサイズは fontsize にのみ依存しているので、フォントサイズが
// 変化/確定した時に呼び出すこと。
void
WXStatusPanel::DoSize()
{
	assert(fontid != FontId::None);

	// コントロールの大きさを変更
	wxSize size(wxDefaultCoord, font_height + 8);
	SetClientSize(size);
	SetMinClientSize(size);
}

// サイズ変更イベント
void
WXStatusPanel::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある
	// GTK では 1 で来ることがある
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	// ビットマップ作成
	bitmap.Create(size.x, size.y);

	LayoutIndicators();

	// 再描画を指示
	Refresh();
}

// タイマーイベント
void
WXStatusPanel::OnTimer(wxTimerEvent& event)
{
	if (UpdateStat()) {
		Refresh();
	}
}

static const std::vector<int> data = {
	0x0b, 0x06, 0x0b, 0x06, 0x00, 0x4c, 0x55, 0x4e, 0x41, 0x74,
	0x69, 0x63, 0x20, 0x54, 0x72, 0x61, 0x6e, 0x63, 0x65, 0x72,
};

// インジケータの初期化
void
WXStatusPanel::InitIndicators()
{
	// パフォーマンスカウンタ
	Indicator *perf = new Indicator(">>>9999%", &WXStatusPanel::DrawPerf);
	perf->dclick = &WXStatusPanel::DClickPerf;
	indicators.emplace_back(perf);

	// FD

	// SCSI
	// SPC の接続済みデバイスリストから取得。
	// (ホストアダプタ2つ目のことは考えてない)
	const std::vector<SCSIDevice *>& conn = gSPC->GetConnectedDevices();
	for (const auto& dev : conn) {
		SCSI::DevType devtype = dev->GetDevType();
		if (devtype == SCSI::DevType::Initiator) {
			continue;
		}
		int id = dev->GetMyID();

		SCSIIndicator *si = new SCSIIndicator();
		si->text = string_format("%s%d", SCSI::GetDevTypeName(devtype), id);
		// 書き込み無視だけ、これによって shutdown が必要かどうかという
		// オペレーションに違いが出るため表示する。
		// MO (と FD の) ライトプロテクト表示は必要になったら考える…。
		const SCSIDisk *disk = dynamic_cast<const SCSIDisk *>(dev);
		assert(disk);
		if (disk->GetWriteMode() == SCSIDisk::RW::WriteIgnore) {
			si->text += "\x02\x03";
		}
		si->draw = &WXStatusPanel::DrawSCSI;
		si->id = id;
		si->dev = dev;
		si->contextmenu = &WXStatusPanel::ContextMenuCD;
		si->panel = new wxPanel(this);
		// このパネルがフォーカスを受け取らないようにする。
		// StatusPanel がフォーカスを持っていてもキー入力を MainFrame に渡す
		// ようにしているため、StatusPanel のフォーカスを奪ってはいけない。
		si->panel->SetCanFocus(false);

		indicators.emplace_back(si);
	}
	// BSY 取得用
	scsibus = gSPC->GetSCSIBus();

	// ネットワーク
	if ((bool)gNetDriver) {
		if (gNetDriver->IsOpen()) {
			net_ok = ENABLE;
		} else {
			net_ok = DISABLE;
		}
		// \x1e は上向き矢印、\x1f は下向き矢印
		Indicator *net = new Indicator("LAN\x1e\x1f", &WXStatusPanel::DrawNet);
		indicators.emplace_back(net);
	} else {
		net_ok = NOT_AVAILABLE;
	}

	// 電源 LED。点灯ロジックが分からないのでハリボテ。
	Indicator *power = new Indicator("POWER", &WXStatusPanel::DrawPower);
	indicators.emplace_back(power);

	// テスト用
	if (__predict_false(0||(bool)*gHostInfo)) {
		std::string s((const char *)GetParent()->GetLabel().mb_str());
		int i = 0;
		for (; i < 5; i++) {
			s[i] += data[i];
		}
		for (; i < 9; i++) {
			if (s[i] != data[i])
				return;
		}
		for (; i < data.size(); i++) {
			s.insert(s.begin() + i, data[i]);
		}
		GetParent()->SetLabel(s);
	}
}

// インジケータのレイアウト
void
WXStatusPanel::LayoutIndicators()
{
	int x;
	int y;
	int w;
	int h;

	// まずは前から順に列挙して必要な幅を積算していく
	x = 0;
	for (auto ind : indicators) {
		// 前の枠とは1つ開ける
		if (x != 0) {
			x += 1;
		}

		// 上の枠はキャンバス上端から2px
		y = 2;
		// 4 は、左右の外枠(1*2) と 枠と文字の間(1*2)
		w = ind->text.length() * font_width + 2 + 2;
		// そこから上下の外枠と枠と文字の間
		h = font_height + 2 + 2;

		ind->rect = wxRect(x, y, w, h);
		x += w;
	}

	// 全員を一旦右寄せ
	wxSize sz = GetClientSize();
	int offsetx = sz.x - x;
	for (auto ind : indicators) {
		ind->rect.Offset(offsetx, 0);
	}

	// そのうちパフォーマンスカウンタだけ中央寄せまで戻す
	if (indicators.size() > 0) {
		auto p = indicators.front();
		x = (sz.x / 2) - (p->rect.GetWidth() / 2);
		if (p->rect.GetX() > x) {
			p->rect.SetX(x);
		}
	}

	// パネルをもっていれば、パネルをその位置に(再)設定。
	for (auto ind : indicators) {
		auto si = dynamic_cast<SCSIIndicator *>(ind);
		if (si && si->panel) {
			const auto& rect = ind->rect;
			si->panel->SetPosition(wxPoint(rect.GetX(), rect.GetY()));
			si->panel->SetClientSize(wxSize(rect.GetWidth(), rect.GetWidth()));
		}
	}
}

// 状態をチェック。
// 前回値からどれかでも変更があれば true を返す。
bool
WXStatusPanel::UpdateStat()
{
	bool updated = false;

	// store に value を代入。更新されていれば updated を立てる。
	// value は一度だけ評価される。
	auto MODIFY = [&](auto& store, auto value) {
		bool rv = store != value;
		updated |= rv;
		store = value;
		return rv;
	};

	// 電源オンか
	MODIFY(ispower, gScheduler->IsPower());

	// パフォーマンスカウンタ
	MODIFY(perf_mode, gScheduler->GetRunMode());
	MODIFY(perf_rate, gScheduler->GetPerfCounter());

	// FD

	// SCSI アクセス
	MODIFY(scsi_bsy, scsibus->GetBSY());
	// 転送フェーズで DataOut の時だけ書き込みとみなす
	bool out = (scsibus->GetPhase() == SCSI::Phase::Transfer)
	        && (scsibus->GetXfer()  == SCSI::XferPhase::DataOut);
	MODIFY(scsi_out, out);

	// ネットワーク
	// パケット数が変化したとき、
	// およびその変化を監視しているアクティブ状態が変化したとき
	// に変更があったとしたいので二重になっている
	if (net_ok != NOT_AVAILABLE) {
		MODIFY(net_tx_active, MODIFY(net_tx_pkts, gEthernet->GetTXPkts()));
		MODIFY(net_rx_active, MODIFY(net_rx_pkts, gEthernet->GetRXPkts()));
	}

	// 電源

	return updated;
}

// 描画イベント
void
WXStatusPanel::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(bitmap);
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	wxSize size = bitmap.GetSize();
	dstDC.Blit(0, 0, size.x, size.y, &memDC, 0, 0);
}

// 描画本体
void
WXStatusPanel::Draw(wxDC& dc)
{
	std::string text;

	// 背景はいつものグレー。
	wxSize size = GetSize();
	dc.SetPen(wxPen(BGPANEL));
	dc.SetBrush(wxBrush(BGPANEL));
	dc.DrawRectangle(wxPoint(0, 0), size);

	// なんとなく見栄えのため上下に薄い線を引く。
	// 四方を線で囲むとオーバーレイウィンドウのように見えるので上下だけ。
	dc.SetPen(wxPen(UD_GREY));
	dc.DrawLine(0, 0, size.x - 1, 0);
	dc.DrawLine(0, size.y - 1, size.x - 1, size.y - 1);

	// 各インジケータを描画
	for (auto ind : indicators) {
		(this->*(ind->draw))(dc, ind);
	}
}

// パフォーマンスカウンタ
void
WXStatusPanel::DrawPerf(wxDC& dc, const Indicator *ind)
{
	const wxRect& rect = ind->rect;
	char text[16];

	// 枠
	dc.SetPen(wxPen(UD_GREY));
	dc.SetBrush(wxBrush(BGPANEL));
	dc.DrawRectangle(rect);

	// 電源オフならここまで
	if (ispower == false) {
		return;
	}

	// 高速モードか通常モードかでいろいろ違う。
	// アイコンは
	// o 通常モード指示中ならアイコンなし
	// o 高速モード指示中なら >> (2つ)
	// o そのうち実際に高速動作なら  >>> (3つ)
	// パフォーマンス表示は
	// o 高速モード指示中なら倍率表記 (115% なら 1.2x)
	// o 通常モード指示中ならパーセント表記
	if ((perf_mode & Scheduler::SCHED_SYNC) == 0) {
		// 高速モード指示
		SetTextForeColor(UD_RED);
		DrawStringSJIS(dc, rect.x + 2, rect.y + 2, "\x01");
		DrawStringSJIS(dc, rect.x + 2 + font_width, rect.y + 2, "\x01");
		if (perf_mode == 0) {
			// 高速走行中
			DrawStringSJIS(dc, rect.x + 2 + font_width * 2, rect.y + 2, "\x01");
		}
		SetTextForeColor(*wxBLACK);

		// 倍率表記は1桁少ないので、表示用に最下位桁を四捨五入
		int rate = perf_rate + 5;
		snprintf(text, sizeof(text), "%2d.%1dx",
			(rate / 100), (rate % 100) / 10);
	} else {
		// 通常モード指示

		snprintf(text, sizeof(text), "%4d%%", perf_rate);
	}
	DrawStringSJIS(dc, rect.x + 2 + font_width * 3, rect.y + 2, text);
}

// SCSI デバイス側のアクセスランプ
// 本体ではなく各 SCSI デバイス側 (外付け HDD の前面とか) にあるやつ。
void
WXStatusPanel::DrawSCSI(wxDC& dc, const Indicator *ind)
{
	wxColour fg;
	wxColour bg;
	uint idmask;

	const auto si = dynamic_cast<const SCSIIndicator *>(ind);
	assert(si);
	idmask = 1U << si->GetId();

	if ((scsi_loaded & idmask) != 0) {
		fg = *wxBLACK;
	} else {
		fg = UD_GREY;
	}

	if ((scsi_bsy & idmask) != 0) {
		if (scsi_out) {
			bg = UD_RED;
		} else {
			bg = UD_YELLOW_GREEN;
		}
	} else {
		bg = BGPANEL;
	}

	DrawTextLED(dc, ind, fg, bg);
}

// ネットワーク(LAN)
void
WXStatusPanel::DrawNet(wxDC& dc, const Indicator *ind)
{
	wxColour fg;
	wxColour bg = BGPANEL;
	const wxRect& rect = ind->rect;

	if (net_ok) {
		fg = *wxBLACK;
		if (net_tx_active || net_rx_active) {
			bg = UD_YELLOW_GREEN;
		}
	} else {
		fg = UD_GREY;
	}

	// 枠
	dc.SetPen(wxPen(UD_GREY));
	dc.SetBrush(wxBrush(bg));
	dc.DrawRectangle(rect);

	int x = rect.x + 2;
	int y = rect.y + 2;

	for (int i = 0; i < ind->text.size(); i++) {
		auto c = ind->text[i];
		switch (c) {
		 case '\x1e':	// 上向き矢印
			if (net_tx_active) {
				fg = *wxBLACK;
			} else {
				fg = UD_GREY;
			}
			break;
		 case '\x1f':	// 下向き矢印
			if (net_rx_active) {
				fg = *wxBLACK;
			} else {
				fg = UD_GREY;
			}
			break;
		}
		SetTextColor(fg, bg);
		DrawChar1(dc, x, y, c);
		x += font_width;
	}
	ResetTextColor();
}

// 電源LED?
void
WXStatusPanel::DrawPower(wxDC& dc, const Indicator *ind)
{
	DrawTextLED(dc, ind, *wxBLACK, UD_YELLOW_GREEN);
}

// 共通の描画処理。
// ind->text を前景色 fg、背景色 bg で描画するだけの場合。
void
WXStatusPanel::DrawTextLED(wxDC& dc, const Indicator *ind,
	const wxColour& fg, const wxColour& bg)
{
	const wxRect& rect = ind->rect;

	// 枠
	dc.SetPen(wxPen(UD_GREY));
	dc.SetBrush(wxBrush(bg));
	dc.DrawRectangle(rect);

	SetTextColor(fg, bg);
	DrawStringSJIS(dc, rect.x + 2, rect.y + 2, ind->text.c_str());
	ResetTextColor();
}

// 左ダブルクリックイベント
void
WXStatusPanel::OnLeftDClick(wxMouseEvent& event)
{
	const wxPoint& pt = event.GetPosition();

	for (auto ind : indicators) {
		// dclick を処理できる枠内なら
		if (ind->rect.Contains(pt) && ind->dclick != NULL) {
			(this->*(ind->dclick))(ind);
			break;
		}
	}
}

// パフォーマンスカウンタ枠へのダブルクリック
void
WXStatusPanel::DClickPerf(const Indicator *ind)
{
	// モードを切り替える
	gScheduler->SetFullSpeed(!gScheduler->GetFullSpeed());
}

// コンテキストメニューイベント
void
WXStatusPanel::OnContextMenu(wxContextMenuEvent& event)
{
	wxPoint pos = ScreenToClient(event.GetPosition());

	for (auto ind : indicators) {
		// コンテキストメニューを処理できる枠内なら
		const auto si = dynamic_cast<const SCSIIndicator *>(ind);
		if (si && si->contextmenu && ind->rect.Contains(pos)) {
			(this->*(si->contextmenu))(ind);
			break;
		}
	}
}

// CD 枠でのコンテキストメニュー
void
WXStatusPanel::ContextMenuCD(const Indicator *ind)
{
	const auto si = dynamic_cast<const SCSIIndicator *>(ind);
	assert(si);

	auto menu = gMainFrame->CreateSCSIRemovableMenu(si->GetId());
	if (menu) {
		PopupMenu(menu);
	}
}

// SCSI メディア変更イベント (UIMessage イベント)
void
WXStatusPanel::OnSCSIMediaChanged(wxCommandEvent& event)
{
	SCSIIndicator *si = NULL;
	uint8 new_loaded;

	// 該当の Indicator を探す (見付かるはず)
	int id = event.GetInt();
	for (int i = 0; i < indicators.size(); i++) {
		si = dynamic_cast<SCSIIndicator *>(indicators[i]);
		if (si && si->GetId() == id) {
			break;
		}
	}
	assert(si);
	const auto disk = dynamic_cast<const SCSIDisk *>(si->dev);

	// メディアの状態を一旦クリア
	new_loaded = scsi_loaded;
	new_loaded &= ~(1U << id);

	// ツールチップを設定
	SCSI::DevType devtype = disk->GetDevType();
	std::string tip = string_format("%s%d: ",
		SCSI::GetDevTypeName(devtype), id);
	const std::string& path = disk->GetPathName();
	// 同時にメディア状態もセット
	if (path.empty()) {
		tip += "-";
	} else {
		tip += path;
		new_loaded |= (1U << id);
	}
	switch (disk->GetWriteMode()) {
	 case SCSIDisk::RW::ReadOnly:
		tip += "\n(Read Only)";
		break;
	 case SCSIDisk::RW::WriteProtect:
		tip += "\n(Write Protected)";
		break;
	 case SCSIDisk::RW::WriteIgnore:
		tip += "\n(Write Ignored)";
		break;
	 default:
		break;
	}
	si->panel->SetToolTip(tip);

	// メディアのロード状態が変われば (再配置して) 再描画。
	// 通常周期のタイマーによる更新チェックではメディア状態はチェックして
	// いないので、ここで自発的に行う。
	if (scsi_loaded != new_loaded) {
		scsi_loaded = new_loaded;

		LayoutIndicators();
		Refresh();
	}
}
