//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "wxnono.h"
#include "wxbitmapbuf.h"
#include "wxtextpanel.h"
#include "scsibus.h"

class Indicator;
class WXStatusPanel;

// インジケータ情報 (基本クラス)
// クラス継承の構造をとっているが、基本的には (関数ポインタを持つ) 構造体
// テイストで扱うほうがここでは都合がよさそう。
class Indicator
{
 protected:
	using dispfunc_t = void (WXStatusPanel::*)(wxDC&, const Indicator *);
	using eventfunc_t = void (WXStatusPanel::*)(const Indicator *);
 public:
	Indicator() { }
	Indicator(const std::string& text_, dispfunc_t draw_) {
		text = text_;
		draw = draw_;
	}
	virtual ~Indicator() = default;

	// 表示する文字列。
	// 描画関数はこの文字列を使っても使わなくてもよいが、InitIndicators() が
	// この文字数に応じた大きさの枠を用意するため、使わない場合でも文字数
	// だけは揃えておくこと。
	std::string text {};

	// 描画関数 (必須)
	dispfunc_t draw {};

	// ダブルクリックの処理関数 (必要なら)
	eventfunc_t dclick {};

	// 枠の位置と大きさは panel.GetRect() で取得できる (実行時に計算する)
	wxRect rect {};
};

// インジケータ情報 (SCSI)
class SCSIIndicator : public Indicator
{
	using inherited = Indicator;
 public:
	// SCSI ID
	int GetId() const { return id; }
	int id {};

	// 対応するデバイス
	SCSIDevice *dev {};

	// ToolTip を表示するためのダミーパネル
	wxPanel *panel {};

	// コンテキストメニューの処理関数 (必要なら)
	eventfunc_t contextmenu {};
};

// ステータスパネル
class WXStatusPanel : public WXTextPanel
{
	using inherited = WXTextPanel;
 public:
	WXStatusPanel(wxWindow *parent);
	virtual ~WXStatusPanel() override;

	void SetFontSize(FontId) override;

 private:
	void OnClose(wxCloseEvent&);
	void OnShow(wxShowEvent&);
	void OnSize(wxSizeEvent&);
	void OnTimer(wxTimerEvent&);
	void OnPaint(wxPaintEvent&);
	void OnLeftDClick(wxMouseEvent&);
	void OnContextMenu(wxContextMenuEvent&);
	void OnSCSIMediaChanged(wxCommandEvent&);

	void DoSize();
	bool UpdateStat();
	void Draw(wxDC&);
	void InitIndicators();
	void LayoutIndicators();
	void DrawPerf(wxDC&, const Indicator *);
	void DrawSCSI(wxDC&, const Indicator *);
	void DrawNet(wxDC&, const Indicator *);
	void DrawPower(wxDC&, const Indicator *);
	void DrawTextLED(wxDC&, const Indicator *,
		const wxColour& fg, const wxColour& bg);
	void DClickPerf(const Indicator *);
	void ContextMenuCD(const Indicator *);

	// コントロール全域のビットマップ
	WXBitmapBuf bitmap {};

	// インジケータ情報
	std::vector<Indicator*> indicators {};

	// インジケータ表示用
	bool ispower {};
	int perf_mode {};
	int perf_rate {};
	std::shared_ptr<SCSIBus> scsibus {};
	uint8 scsi_bsy {};
	bool scsi_out {};
	uint8 scsi_loaded {};
	enum {
		NOT_AVAILABLE = -1,	// 存在しない (Nereid なし X68k とか)
		DISABLE = 0,		// 存在してホスト側が無効 (driver = none)
		ENABLE = 1,			// 存在してホスト側が有効
	} net_ok {};
	uint64 net_tx_pkts {};
	uint64 net_rx_pkts {};
	bool net_tx_active {};
	bool net_rx_active {};

	// タイマー
	wxTimer timer {};

	// イベントテーブル
	wxDECLARE_EVENT_TABLE();
};

extern WXStatusPanel *gStatusPanel;
